#!/usr/bin/python

#
# (C) Copyright 2005, 2018-2020 Micro Focus or one of its affiliates.
# All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General Public
# License as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, contact Novell, Inc.
#
# To contact Novell about this file by physical or electronic mail,
# you may find current contact information at www.novell.com
#
#
#  Author:  Novell, Inc.
#  Workfile:  clstrlib.py
#  Contents:  Methods used to install clusters, plus
#             Userspace implementation of CLSTRLIB eDir access
#

import sys
import ldap
import dsml
import os
import time
import subprocess
import os.path
import struct
import string
import re
import ncstempl
import syslog
import xml.sax
from xml.sax.handler import *
import xml.dom.minidom

clstrlibPath = "/var/opt/novell/ncs"
allNodes = "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"

#clstrlib/src/conf.c
CLSTRCONF_SCRIPTS_BUFFER_LEN = (1024 * 1024)

"""
 For writing the data to the script files, there is no restriction as such.
 But for reading the same, a buffer of 1MB(clstrconf_scripts_buffer) is allocated.
 Hence, while writing itself only 1MB of data is allowed for sp3 and above,
 3200 bytes for the rest to maintain backward compatibility.
"""
EDIR_ATTR_LIMIT_OLD = 3200

#############################################################################
#
# Class: ClstrlibError
#
#   Purpose:
#
class ClstrlibError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr( self.value )

#############################################################################
#
# Function: daemonize
#
#   Purpose: turn into a daemon
#
def daemonize():
    # do the UNIX double-fork magic, see Stevens' "Advanced
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            os._exit(0)
    except OSError, e:
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror)
        syslog.syslog(syslog.LOG_DEBUG, "fork #1 failed: %d (%s)" % (e.errno, e.strerror))
        return 1

    # decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0022)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent, print daemon PID to syslog
            # syslog.syslog(syslog.LOG_DEBUG, "PID=%d" % pid)
            os._exit(0)
    except OSError, e:
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror)
        syslog.syslog(syslog.LOG_DEBUG, "fork #2 failed: %d (%s)" % (e.errno, e.strerror))
        os._exit(1)

    # Close all open files.  Try the system configuration variable, SC_OPEN_MAX,
    # for the maximum number of open files to close.  If it doesn't exist, use
    # the default value.
    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
    except (AttributeError, ValueError):
        maxfd = 256       # default maximum

    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError:   # ERROR (ignore)
            pass
    try:
        sys.stdin = open( "/dev/null", "r" )
        sys.stdout = open( "/dev/null", "w" )
        sys.stderr = open( "/dev/null", "w" )
    except:
        pass

    return 0

#############################################################################
#
# Function: eDirToLdapDn
#
#   Purpose:
#
def eDirToLdapDn(eDirDN):

    if (eDirDN.startswith("cn=") == True):
        return eDirDN

    if (eDirDN.startswith(".T") == True):
        components = string.split(eDirDN, ".")
        components.reverse()
    else:
        components = string.split(eDirDN, ".")

    index = 1
    ldapDn = ""

    for component in components:

        if (component != ""):

            parts = string.split(component, "=")

            if (len(parts) == 2):
                if (parts[0] != "T"):
                    if (ldapDn != ""):
                        ldapDn += ","
                    ldapDn += string.lower(parts[0]) + "=" + parts[1]

            else:
                if (ldapDn == ""):
                    ldapDn = "cn=" + parts[0]
                elif (index == len(components)):
                    ldapDn += ",o=" + parts[0]
                else:
                    ldapDn += ",ou=" + parts[0]

        index = index + 1

    return ldapDn

#############################################################################
#
# Function: containerFromLdapDn
#
#   Purpose:
#
def containerFromLdapDn(ldapDn):

    containerDn = ""

    for component in string.split(ldapDn, ",")[1:]:

        if (containerDn == ""):
            containerDn = component

        else:
            containerDn += "," + component

    return containerDn

#############################################################################
#
# Function: objectNameFromLdapDn
#
#   Purpose:
#
def objectNameFromLdapDn(ldapDn):

    cn = string.split(ldapDn, ",")[0]
    return string.split(cn, "=")[1]

#############################################################################
#
# Function: filterSearchResultByObjectClass
#
#   Purpose: Filter an LDAP search result list and return a new list
#            containing objects that match the specified objectClass.
#            objectClass can be a regular expression. For example:
#            "nCS(.*?)Resource(.*?)" will match "nCSClusterResource",
#            "nCSVolumeResource" or "nCSResourceTemplate"
#
def filterSearchResultByObjectClass(res, filter):

    out = []
    reg = re.compile(filter, re.IGNORECASE)

    for r in res:
        if r[1].has_key("objectClass"):
            for oc in r[1]["objectClass"]:
                mo = reg.search(oc)
                if mo:
                    out.append(r)
                    break;

    return out

#############################################################################
#
# Function: createCluster
#
#   Purpose:
#
def createCluster(ld, containerDn, clusterName, ipAddress, sharedDiskFlag=False):

    clusterDn = "cn=" + clusterName + "," + containerDn
    ip = string.split(ipAddress, ".")
    ipBytes = struct.pack("BBBB", int(ip[0]), int(ip[1]), int(ip[2]), int(ip[3]))

    attrs = []
    attrs.append( ("cn", clusterName) )
    attrs.append( ("objectClass", "nCSNetWareCluster") )
    attrs.append( ("nCSGIPCMaxRetransmits", "30") )
    attrs.append( ("nCSCRMQuorum", "2") )
    attrs.append( ("nCSCRMQuorumTimeout", "60") )
    attrs.append( ("nCSGIPCHeartbeat", "1") )
    attrs.append( ("nCSRevision", "282") )
    attrs.append( ("nCSGIPCSlaveWatchdog", "8") )
    attrs.append( ("nCSGIPCTolerance", "8") )
    attrs.append( ("nCSPortNumber", "7023") )
    attrs.append( ("nCSGIPCMasterWatchdog", "1") )
    attrs.append( ("nCSCRMResourcePriority", "REVISION=0\r\nNUMRESOURCES=1\r\nRESOURCE=Master_IP_Address_Resource=65535\r\n") )
    attrs.append( ("nCSNetworkAddress", "1#" + ipBytes) )
    attrs.append( ("networkAddress", [ "8#\x02\x0c" + ipBytes, "9#\x02\x0c" + ipBytes ]) )
    attrs.append( ("nCSSharedDiskFlag", str(sharedDiskFlag).upper()) )
    attrs.append( ("nCSEmailFilter", "11") )
    attrs.append( ("nCSEmailAddresses", [ "cluster@novell.com" ]) )

    ld.add_s(clusterDn, attrs)

    ncstempl.updateResourceTemplates(ld, clusterDn)
    createMasterIPAddressResource(ld, clusterDn, ipAddress)

#############################################################################
#
# Function: createNode
#
#   Purpose:
#
def createNode(ld, clusterDn, nodeName, serverDn, nodeNumber, ipBytes):

    attrs = []
    attrs.append( ("cn", nodeName) )
    attrs.append( ("objectClass", "nCSNCPServer") )
    if serverDn:
        attrs.append( ("nCSNCPServer", serverDn) )
    attrs.append( ("nCSGIPCNodeNumber", nodeNumber) )
    attrs.append( ("nCSNetworkAddress", "1#" + ipBytes) )
    attrs.append( ("nCSNodeIsolationScript", "panic") )

    ld.add_s("cn=" + nodeName + "," + clusterDn, attrs)

#############################################################################
#
# Function: deleteServerFromCluster
#
#   Purpose:
#
def deleteServerFromCluster(ld, clusterDn, serverDn ):
    nodeMarked = 0
    removeCluster = 0
    serverName = objectNameFromLdapDn( serverDn )

    try:
        res = ld.search_s( serverDn, ldap.SCOPE_BASE, "cn=" + serverName, [ "cn", "nCSNetWareCluster" ] )
        serverDn = res[0][0]
        serverName = res[0][1]["cn"][0]
        syslog.syslog( syslog.LOG_DEBUG, "serverDn=" + serverDn )
        syslog.syslog( syslog.LOG_DEBUG, "serverName=" + serverName )
        if res[0][1].has_key( "nCSNetWareCluster" ):
            nodeMarked = 1
            syslog.syslog( syslog.LOG_DEBUG, "Need to removed NCS attribute." )
    except ldap.LDAPError, e:
        syslog.syslog( syslog.LOG_ERR, "Failed to access server for NCS related attributes: " + str( e ))
        raise ClstrlibError ( "Failed to access server '" + serverDn + "' for NCS related attributes" )

    try:
        res = ld.search_s( clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer", ["cn"] )
        syslog.syslog( syslog.LOG_DEBUG, "Nodes='%s'" % str( res ))
        syslog.syslog( syslog.LOG_DEBUG, "Number of nodes = %d" % len( res ))
        if 0 == len( res ):
            removeCluster = 1
        for node in res:
            if serverName == node[1]["cn"][0]:
                 try:
                     ld.delete_s( node[0] )
                     syslog.syslog( syslog.LOG_INFO, "Clustered node '" + node[0] + "' removed.")
                     if 1 == len( res ):
                         removeCluster = 1
                 except ldap.LDAPError, e:
                     syslog.syslog( syslog.LOG_ERR, "Failed to remove clustered node '" + node[0] + "': " + str( e ))
                     raise ClstrlibError ( "Failed to remove '" + node[0] + "' from the cluster." )
                 break
    except ldap.LDAPError, e:
        if( 0 == nodeMarked ):
            pass
        else:
            syslog.syslog( syslog.LOG_ERR, "Failed to access the cluster object: " + str( e ))
            raise ClstrlibError ( "Failed to access the cluster object." )

    syslog.syslog( syslog.LOG_DEBUG, "deleteServerFromCluster: nodeMarked=%d, removeCluster=%d" % (nodeMarked, removeCluster ))

    if 0 != nodeMarked:
        attrs = []
        attrs.append(( ldap.MOD_DELETE, "nCSNetWareCluster", clusterDn ))
        try:
            ld.modify_s( serverDn, attrs )
            syslog.syslog( syslog.LOG_INFO, "Attribute '" + str(attrs) + "' removed from '" + serverDn + "'" )
        except ldap.LDAPError, e:
            syslog.syslog( syslog.LOG_ERR, "Failed to remove attribute '" + str(attrs) + "' removed from '" + serverDn + "': " + str( e ))
            raise ClstrlibError( "Failed to remove attribute from server." )

    if 0 != removeCluster:
        try:
            res = ld.search_s( clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=*", ["cn"] )
            syslog.syslog( syslog.LOG_DEBUG, "objects='%s'" % str( res ))
            for node in res:
                try:
                    ld.delete_s( node[0] )
                    syslog.syslog( syslog.LOG_INFO, "Object '" + node[0] + "' removed.")
                except:
                    pass
            try:
                ld.delete_s( clusterDn )
                syslog.syslog( syslog.LOG_INFO, "Cluster '" + clusterDn + "' removed.")
                try:
                    ld.delete_s( "cn=" + objectNameFromLdapDn( clusterDn ) + "_MGT_GRP," + containerFromLdapDn( clusterDn ))
                    syslog.syslog( syslog.LOG_INFO, "Cluster management group removed.")
                except:
                    pass
            except ldap.LDAPError, e:
                syslog.syslog( syslog.LOG_ERR, "Failed to remove cluster '" + clusterDn + "': " + str( e ))
                raise ClstrlibError( "Failed to delete cluster." )
        except:
            pass


#############################################################################
#
# Function: addServerToCluster
#
#   Purpose:
#
def addServerToCluster(ld, clusterDn, serverDn, nodeName=None, ipAddress=None):

    # Check cluster revision
    res = ld.search_s(clusterDn, ldap.SCOPE_BASE)
    revision = int(res[0][1]["nCSRevision"][0])
    if (revision < 261):
        raise ClstrlibError ( "cannot add server to mixed cluster: revision too old" )
    elif (revision < 270):
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSRevision", "275") )
        attrs.append( (ldap.MOD_ADD, "nCSSharedDiskFlag", "TRUE") )
        ld.modify_s(clusterDn, attrs)
    elif (revision < 275):
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSRevision", "280") )
        attrs.append( (ldap.MOD_ADD, "nCSSharedDiskFlag", "TRUE") )
        ld.modify_s(clusterDn, attrs)

    # Generate the next free node number and number of nodes
    numNodes = 0
    nodeNumberList = range(33)
    res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer", [ "nCSGIPCNodeNumber" ] )
    for node in res:
        if not node[1].has_key( "nCSGIPCNodeNumber" ):
            continue;
        numNodes = numNodes + 1
        nodeNumberList.remove(int(node[1]["nCSGIPCNodeNumber"][0]))
    nodeNumber = str(nodeNumberList[0])
    if (nodeNumber == "32"):
        raise ClstrlibError ( "max number of nodes already configured in cluster" )
    numNodes = numNodes + 1

    # If ipAddress is specified, convert into ipBytes
    if ipAddress:
        ip = string.split(ipAddress, ".")
        ipBytes = struct.pack("BBBB", int(ip[0]), int(ip[1]), int(ip[2]), int(ip[3]))
    else:
        ipBytes = None

    # Get nodeName and ipBytes from serverDn if specified
    if serverDn:
        # Inspect the NCP Server given by serverDn
        res = ld.search_s(serverDn, ldap.SCOPE_BASE)

        # Check serverDn already linked
        if res[0][1].has_key("nCSNetWareCluster"):
            raise ClstrlibError ( serverDn + " already configured in cluster" )

        # Get the NCP Server's common name
        nodeName = res[0][1]["cn"][0]

        # Get the NCP Server's network address
        if not ipBytes:
            for networkAddress in res[0][1]["networkAddress"]:
                if (networkAddress[:2] == "8#"):
                    ipBytes = networkAddress[4:]
                    break

    # For mixed clusters, use GIPC config if present
    res = ld.search_s(clusterDn, ldap.SCOPE_BASE)
    if (res[0][1].has_key("nCSGIPCConfig") and ipBytes):
        nwGipcConfig = res[0][1]["nCSGIPCConfig"][0]
        nodeIp = struct.unpack("BBBB", ipBytes)
        nodeIpAddress = str(nodeIp[0]) + "." + str(nodeIp[1]) + "." + str(nodeIp[2]) + "." + str(nodeIp[3])
        left = nwGipcConfig.find(nodeIpAddress)
        if (left != -1):
            right = nwGipcConfig[left:].find("\n")
            nodeConfig = nwGipcConfig[left:][:right].strip().split()
            if (nodeConfig[0] == nodeIpAddress):
                nwNodeNumber = nodeConfig[1]
                try:
                    i = nodeNumberList.index(int(nwNodeNumber))
                    nodeNumber = nwNodeNumber
                except:
                    pass

    # Create cluster node if nodeName and ipBytes specified
    if (nodeName and ipBytes):
        createNode(ld, clusterDn, nodeName, serverDn, nodeNumber, ipBytes)
    else:
        raise ClstrlibError ( "nodeName and ipBytes must be specified" )

    if serverDn:
        # Link serverDn to clusterDn
        attrs = []
        attrs.append( (ldap.MOD_ADD, "nCSNetWareCluster", clusterDn) )
        try:
            ld.modify_s(serverDn, attrs)
        except:
            pass

    # Update quorum
    attrs = []
    attrs.append( (ldap.MOD_REPLACE, "nCSCRMQuorum", str(numNodes)) )
    try:
        ld.modify_s(clusterDn, attrs)
    except:
        pass

    return None

#############################################################################
#
# Function: addUserToNCSGroup
#
# Returns: 0 or error code
#
# Comments:
#
def addUserToNCSGroup( ld, clusterDn, member ):
    ncsGroup = "cn=" + objectNameFromLdapDn( clusterDn ) + "_MGT_GRP," + containerFromLdapDn( clusterDn )
    syslog.syslog( syslog.LOG_DEBUG, "ncsGroup=" + ncsGroup )

    retMember = syncObject( ld, member )
    if not retMember:
        syslog.syslog( syslog.LOG_DEBUG, "user '" + member + "' not found." )
        return -1
    member = retMember

    if not syncObject( ld, ncsGroup ):
        syslog.syslog( syslog.LOG_DEBUG, "group '" + ncsGroup + "' not found." )

        attrs = []
        attrs.append(( "cn", objectNameFromLdapDn( ncsGroup )))
        attrs.append(( "objectClass", "groupOfNames" ))
        try:
            ld.add_s( ncsGroup, attrs )
        except ldap.LDAPError, e:
            syslog.syslog( syslog.LOG_ERR, "failed to create group '" + ncsGroup + "'\n" + str( e ))
            sys.stderr.write( "Error: failed to create group '" + ncsGroup + "'\n" + str( e ))
            return -2

    search_domains = set()
    search_domains.add( containerFromLdapDn(clusterDn))

    try:
        res = ld.search_s( clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer", [ "nCSNCPServer" ] )
        for node in res:
            if node[1].has_key( "nCSNCPServer" ):
                search_domains.add( containerFromLdapDn( node[1]["nCSNCPServer"][0] ))
    except:
        pass

    for domain in search_domains:
        syslog.syslog( syslog.LOG_DEBUG, "Searching ACL for " + domain )
        try:
            attrs = []
            res = ld.search_s( domain, ldap.SCOPE_BASE )
            newACL = "31#subtree#" + ncsGroup + "#[Entry Rights]"
            if 0 < len( res ) and 1 < len( res[0] ) and res[0][1].has_key( "ACL" ):
                try:
                    i = res[0][1]["ACL"].index( newACL )
                except:
                    i = -1

                if -1 == i:
                    syslog.syslog( syslog.LOG_DEBUG, "append ACL: " + newACL )
                    res[0][1]["ACL"].append( newACL )
                    attrs.append(( ldap.MOD_REPLACE, "ACL", res[0][1]["ACL"] ))
                else:
                    syslog.syslog( syslog.LOG_DEBUG, "ACL exist: " + newACL )
            else:
                syslog.syslog( syslog.LOG_DEBUG, "add ACL: " + newACL )
                attrs.append((  ldap.MOD_ADD, "ACL", [ newACL ] ))

            syslog.syslog( syslog.LOG_DEBUG, "items in attribute array: " + str( len( attrs )))
            if attrs :
                try:
                    ld.modify_s( domain, attrs )
                except ldap.LDAPError, e:
                    syslog.syslog( syslog.LOG_INFO, "failed to add ACL for group '" + ncsGroup + "' to object '" + domain + "'.\n" + str( e ))
                    sys.stderr.write( "Warning: failed to add ACL for group '" + ncsGroup + "' to object '" + domain + "'.\n" + str( e ))

        except ldap.LDAPError, e:
            syslog.syslog( syslog.LOG_INFO, "failed to access '" + domain + "'\n" + str( e ))
            sys.stderr.write( "Warning: failed to access '" + domain + "'\n" + str( e ))

    try:
        res = ld.search_s( ncsGroup, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn( ncsGroup ), [ "uniqueMember", "equivalentToMe" ] )
    except ldap.LDAPError, e:
        syslog.syslog( syslog.LOG_INFO, "failed to access group '" + ncsGroup + "'\n" + str( e ))
        sys.stderr.write( "Warning: failed to access group '" + ncsGroup + "'\n" + str( e ))

    attrs = []
    if 0 < len( res ) and 1 < len( res[0] ) and res[0][1].has_key( "uniqueMember" ):
        try:
            i =  res[0][1]["uniqueMember"].index( member )
        except:
            i = -1

        if -1 == i:
            value = res[0][1]["uniqueMember"]
            value.append( member )
            attrs.append(( ldap.MOD_REPLACE, "uniqueMember", value ))
    else:
        attrs.append(( ldap.MOD_ADD, "uniqueMember", [ member ] ))

    if 0 < len( res ) and 1 < len( res[0] ) and res[0][1].has_key( "equivalentToMe" ):
        try:
            i =  res[0][1]["equivalentToMe"].index( member )
        except:
            i = -1

        if -1 == i:
            value = res[0][1]["equivalentToMe"]
            value.append( member )
            attrs.append(( ldap.MOD_REPLACE, "equivalentToMe", value ))
    else:
        attrs.append(( ldap.MOD_ADD, "equivalentToMe", [ member ] ))

    syslog.syslog( syslog.LOG_INFO, "attributes to modify: " + str(attrs))
    if attrs:
        try:
            ld.modify_s( ncsGroup, attrs )
        except ldap.LDAPError, e:
            syslog.syslog( syslog.LOG_ERR, "failed to modify group '" + ncsGroup + "'\n" + str( e ))
            sys.stderr.write( "Error: failed to modify group '" + ncsGroup + "'\n" + str( e ))
            return -3

    try:
        res = ld.search_s( member, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn( member ), [ "groupMembership" ] )
    except ldap.LDAPError, e:
        syslog.syslog( syslog.LOG_INFO, "failed to access user '" + member + "'\n" + str( e ))
        sys.stderr.write( "Warning: failed to access user '" + member + "'\n" + str( e ))

    attrs = []
    if 0 < len( res ) and 1 < len( res[0] ) and res[0][1].has_key( "groupMembership" ):
        try:
            i =  res[0][1]["groupMembership"].index( ncsGroup )
        except:
            i = -1

        if -1 == i:
            value = res[0][1]["groupMembership"]
            value.append( ncsGroup )
            attrs.append(( ldap.MOD_REPLACE, "groupMembership", value ))
    else:
        attrs.append(( ldap.MOD_ADD, "groupMembership", [ ncsGroup ] ))

    if attrs:
        try:
            ld.modify_s( member, attrs )
        except ldap.LDAPError, e:
            syslog.syslog( syslog.LOG_ERR, "failed to add user '" + member + "' to group\n" + str( e ))
            sys.stderr.write( "Error: failed to add user '" + member + "' to group\n" + str( e ))
            return -4

    try:
        res = ld.search_s( member, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn( member ), [ "securityEquals" ] )
        syslog.syslog( syslog.LOG_INFO, "securityEquals: " + str(res))
    except ldap.LDAPError, e:
        syslog.syslog( syslog.LOG_INFO, "failed to access user '" + member + "' for securityEquals\n" + str( e ))
        sys.stderr.write( "Warning: failed to access user '" + member + "' for securityEquals\n" + str( e ))

    attrs = []
    if 0 < len( res ) and 1 < len( res[0] ) and res[0][1].has_key( "securityEquals" ):
        try:
            i =  res[0][1]["securityEquals"].index( ncsGroup )
        except:
            i = -1

        if -1 == i:
            value = res[0][1]["securityEquals"]
            value.append( ncsGroup )
            attrs.append(( ldap.MOD_REPLACE, "securityEquals", value ))
    else:
        attrs.append(( ldap.MOD_ADD, "securityEquals", [ ncsGroup ] ))

    syslog.syslog( syslog.LOG_INFO, "securityEquals attributes to modify: " + str(attrs))
    if attrs:
        try:
            ld.modify_s( member, attrs )
        except ldap.LDAPError, e:
            syslog.syslog( syslog.LOG_ERR, "failed to add securityEquals '" + member + "' to group\n" + str( e ))
            sys.stderr.write( "Error: failed to add securityEquals '" + member + "' to group\n" + str( e ))
            return -5

    return 0

#############################################################################
#
# Function: createMasterIPAddressResource
#
#   Purpose:
#
def createMasterIPAddressResource(ld, clusterDn, ipAddress):

    loadScript =  "#!/bin/bash\n"
    loadScript += ". /opt/novell/ncs/lib/ncsfuncs\n"
    loadScript += "ignore_error add_secondary_ipaddress " + ipAddress + " -np\n"
    loadScript += "exit 0\n"

    unloadScript =  "#!/bin/bash\n"
    unloadScript += ". /opt/novell/ncs/lib/ncsfuncs\n"
    unloadScript += "ignore_error del_secondary_ipaddress " + ipAddress + "\n"
    unloadScript += "exit 0\n"

    attrs = []
    attrs.append( ("cn", "Master_IP_Address_Resource") )
    attrs.append( ("objectClass", "nCSClusterResource") )
    attrs.append( ("nCSCRMLoadTimeout", "360") )
    attrs.append( ("nCSCRMLoadScript", loadScript) )
    attrs.append( ("nCSCRMUnloadTimeout", "360") )
    attrs.append( ("nCSCRMUnloadScript", unloadScript) )
    attrs.append( ("nCSCRMFailbackMode", "0") )
    attrs.append( ("nCSCRMFailoverMode", "-2147483599") )
    attrs.append( ("nCSCRMIgnoreQuorum", "TRUE") )
    attrs.append( ("nCSRevision", "1") )
    attrs.append( ("nCSCRMPreferredNodes", allNodes) )

    ld.add_s("cn=Master_IP_Address_Resource," + clusterDn, attrs)

#############################################################################
#
# Function: getTreeName
#
#   Purpose: get the name of the edir tree
#
def getTreeName(ld, clusterDn, args ) :
    res = ld.search_s('', ldap.SCOPE_BASE, 'objectclass=*', ['directoryTreeName'])
    try :
        treeName = res [0][1]['directoryTreeName'][0]
    except :
        treeName = None
    return treeName

#############################################################################
#
# Function: createVolumeResource
#
#   Purpose:
#
def createVolumeResource(ld, clusterDn, resourceName, poolName, serverName, ipAddress, advertize ):

    resourceDn = "cn=" + resourceName + "," + clusterDn

    containerDn = containerFromLdapDn(clusterDn)
    # CIFS requires UDN
    treeName = getTreeName(ld, clusterDn, clusterDn )
    serverUDN = ".cn=%s.%s.t=%s." % (serverName,containerDn,treeName)
    serverUDN = serverUDN.replace(',','.')


    loadScript = "#!/bin/bash\n"
    loadScript += ". /opt/novell/ncs/lib/ncsfuncs\n"
    loadScript += "exit_on_error nss /poolact=" + poolName + "\n"
    loadScript += "exit_on_error add_secondary_ipaddress " + ipAddress + "\n"
    loadScript += "exit_on_error ncpcon bind --ncpservername=" + serverName + " --ipaddress=" + ipAddress + "\n"
    if 4 & advertize:
        loadScript += "exit_on_error cluster_afp.sh add " + serverName +  " " + ipAddress + "\n"
    if 2 & advertize: # CIFS
        loadScript += "exit_on_error novcifs --add '--vserver=\"" + serverUDN +  "\"' --ip-addr=" + ipAddress + "\n"
    loadScript += "exit 0\n"

    unloadScript = "#!/bin/bash\n"
    unloadScript += ". /opt/novell/ncs/lib/ncsfuncs\n"
    if 2 & advertize: # CIFS
        unloadScript += "ignore_error novcifs --remove '--vserver=\"" + serverUDN +  "\"' --ip-addr=" + ipAddress + "\n"
    if 4 & advertize:
        unloadScript += "ignore_error cluster_afp.sh del " + serverName +  " " + ipAddress + "\n"
    unloadScript += "ignore_error ncpcon unbind --ncpservername=" + serverName + " --ipaddress=" + ipAddress + "\n"
    unloadScript += "ignore_error del_secondary_ipaddress " + ipAddress + "\n"
    unloadScript += "ignore_error nss /pooldeact=" + poolName + "\n"
    unloadScript += "exit 0\n"

    monitorScript = "#!/bin/bash\n"
    monitorScript += ". /opt/novell/ncs/lib/ncsfuncs\n"
    monitorScript += "exit_on_error status_fs /dev/pool/" + poolName + " /opt/novell/nss/mnt/.pools/" + poolName + " nsspool\n"
    monitorScript += "exit_on_error status_secondary_ipaddress " + ipAddress + "\n"
    # todo check the ncp server status
    if 2 & advertize: # CIFS
        monitorScript += "exit_on_error rcnovell-cifs monitor\n"
    if 4 & advertize:
        monitorScript += "exit_on_error afpstat\n"
    monitorScript += "exit 0\n"


    attrs = []
    attrs.append( ("cn", resourceName) )
    attrs.append( ("objectClass", "nCSVolumeResource") )
    attrs.append( ("nCSCRMLoadTimeout", "360") )
    attrs.append( ("nCSCRMLoadScript", loadScript) )
    attrs.append( ("nCSCRMUnloadTimeout", "360") )
    attrs.append( ("nCSCRMUnloadScript", unloadScript) )
    attrs.append( ("nCSCRMMonitorTimeout", "360") )
    attrs.append( ("nCSCRMMonitorScript", monitorScript) )
    attrs.append( ("nCSCRMFailbackMode", "0") )
    attrs.append( ("nCSCRMFailoverMode", "16") )
    attrs.append( ("nCSCRMIgnoreQuorum", "FALSE") )
    attrs.append( ("nCSRevision", "1") )
    attrs.append( ("nCSCRMPreferredNodes", allNodes) )
    attrs.append( ("nCSCRMMonitorFailurePeriod", "600") )
    attrs.append( ("nCSCRMMonitorFrequency", "60") )
    attrs.append( ("nCSCRMMonitorMaxFailures", "3") )

    ld.add_s(resourceDn, attrs)

    return resourceDn

#############################################################################
#
# Function: generateVolumeId
#
#   Purpose:
#
def generateVolumeId():

    # Allocatable ID range is 254 thru 2 inclusive
    volumeIdList = range(2,255)
    volumeIdList.reverse()

    # Search resource load scripts for "ncpcon mount VOL=ID"
    try:
        results = subprocess.Popen( ["/bin/sh", "-c", "grep -v '^[[:space:]]*#' /var/opt/novell/ncs/*.load | grep -i mount | grep -o '\\<\\w\\+=[[:digit:]]\\+\\>'"], close_fds=True, stdout=subprocess.PIPE).communicate()[0]
    except:
        results = ""

    # No IDs in use, return next free
    if (results == ""):
        return volumeIdList[0]

    for line in results.split("\n"):

        if (line != ""):
            try:
                volumeIdList.remove(int(line.split( "=", 1 )[1]))
            except:
                pass

    # Return next free ID
    if (len(volumeIdList) > 0):
        return volumeIdList[0]

    # No IDs left, return -1
    else:
        return -1

#############################################################################
#
# Function: addVolumeToResourceScripts
#
#   Purpose:
#
def addVolumeToResourceScripts(ld, clusterDn, resourceDn, volumeName, volumeId):

    loadScript = ""
    newLoadScript = ""
    newMonitorScript = ""
    tmpVar = resourceDn[3:]
    tmpPool = tmpVar.split("_SERVER,cn")
    poolName = tmpPool[0]

    mount = "exit_on_error ncpcon mount " + volumeName + "=" + str(volumeId)
    ncp_volume = "exit_on_error ncpcon volume " + volumeName

    try:
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        revision = int(res[0][1]["nCSRevision"][0])
        loadScript = res[0][1]["nCSCRMLoadScript"][0]
        try:
            monitorScript = res[0][1]["nCSCRMMonitorScript"][0]
        except:
            monitorScript = None
    except:
        return -1, -1

    newRevision = revision + 1
    searchWord = "nss /poolact="+poolName
    for line in loadScript.split("\n"):
        newLoadScript += line + "\n"
        if (line.find(searchWord) != -1):
            newLoadScript += mount + "\n"

    attrs = []
    attrs.append( (ldap.MOD_REPLACE, "nCSRevision", str(newRevision)) )
    attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadScript", newLoadScript) )

    if monitorScript:
        for line in monitorScript.split("\n"):
            newMonitorScript += line + "\n"
            if (line.find("exit_on_error status_secondary_ipaddress") != -1):
                newMonitorScript += ncp_volume + "\n"
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorScript", newMonitorScript) )

    try:
        ld.modify_s(resourceDn, attrs)

    except:
        return -2, newRevision

    try:
        updateResourceConfig(ld, clusterDn, objectNameFromLdapDn(resourceDn))

    except:
        return -3, newRevision

    return 0, newRevision

#############################################################################
#
# Function: removeVolumeFromResourceScripts
#   Purpose:
#
def removeVolumeFromResourceScripts(ld, clusterDn, resourceDn, volumeName):

    loadScript = ""
    newLoadScript = ""
    newMonitorScript = ""

    ncp_volume = "exit_on_error ncpcon volume " + volumeName

    try:
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        revision = int(res[0][1]["nCSRevision"][0])
        loadScript = res[0][1]["nCSCRMLoadScript"][0]
        try:
            monitorScript = res[0][1]["nCSCRMMonitorScript"][0]
        except:
            monitorScript = None

    except:
        return -1, -1

    newRevision = revision + 1

    for line in loadScript.split("\n"):

        if (line.find("ncpcon mount " + volumeName + "=" ) == -1):
            newLoadScript += line + "\n"

    attrs = []
    attrs.append( (ldap.MOD_REPLACE, "nCSRevision", str(newRevision)) )
    attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadScript", newLoadScript) )

    if monitorScript:
        for line in monitorScript.split("\n"):
            if (line.find(ncp_volume) == -1):
                newMonitorScript += line + "\n"
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorScript", newMonitorScript) )

    try:
        ld.modify_s(resourceDn, attrs)

    except:
        return -2, newRevision

    try:
        updateResourceConfig(ld, clusterDn, objectNameFromLdapDn(resourceDn))

    except:
        return -3, newRevision

    return 0, newRevision

#############################################################################
#
# Function: generateMonitorScript
#   Purpose:
#
def generateMonitorScript( load_script ):
    monitorScript = "#!/bin/bash\n. /opt/novell/ncs/lib/ncsfuncs\n"
    p = re.compile( r"^\s*exit_on_error\s+nss\s+.+$", re.M )
    lines = p.findall( load_script )
    for line in lines:
        pool_name=""
        p = re.compile( r"/poolact=(\w+)" )
        option = p.search( line )
        if option:
            pool_name=option.group(1)

        if pool_name:
            monitorScript += "exit_on_error status_fs /dev/pool/" + pool_name + " /opt/novell/nss/mnt/.pools/" + pool_name + " nsspool\n"

    p = re.compile( r"^\s*exit_on_error\s+add_secondary_ipaddress\s+\d+\.\d+\.\d+\.\d+|^\s*exit_on_error\s+ip\s+addr\s+.+$", re.M )
    lines = p.findall( load_script )
    for line in lines:
        ip_addr=""
        p = re.compile( r"\d+\.\d+\.\d+\.\d+" )
        option = p.search( line )
        if option:
            ip_addr=option.group()

        if ip_addr:
            monitorScript += "exit_on_error status_secondary_ipaddress " + ip_addr + "\n"

    p = re.compile( r"^\s*exit_on_error\s+ncpcon\s+mount\s+.*$", re.M )
    lines = p.findall( load_script )
    for line in lines:
        volume_name = ""
        p = re.compile( r"(\w+)=\d+" )
        option = p.search( line )
        if option:
            volume_name=option.group(1)
        else:
            p = re.compile( r"shadowvolume=(\w+)", re.I )
            option = p.search( line )
            if option:
                volume_name=option.group(1)

        if volume_name:
            monitorScript += "exit_on_error ncpcon volume " + volume_name + "\n"

    p = re.compile( r"^\s*exit_on_error\s+cluster_afp.sh\s+add\s+.*$", re.M )
    if p.search( load_script ):
        monitorScript += "exit_on_error afpstat\n"

    p = re.compile( r"^\s*exit_on_error\s+novcifs\s+--add\s+.*$", re.M )
    if p.search( load_script ):
        monitorScript += "exit_on_error rcnovell-cifs monitor\n"

    monitorScript += "exit 0\n"
    return monitorScript

#############################################################################
#
# Function: translateNetWareScript
#   Purpose: Translates a load/unload script with NetWare commands to the
#               Linux equivalent.  Returns a two element list.  The first
#               element in the list is a boolean specifying if the script
#               changed.  The second element is the script itself.
#
def translateNetWareScript (script, syntax_file="/var/opt/novell/ncs/customized_translation_syntax" ):

    changed = False
    xforms = []
    newScript = []

    # Build the list of translations.  Any changes to this list require
    # the resource daemon to be restarted.  The format of the tuple is as
    # follows:
    # 'Action', 'Search', 'Replace', 'Position'
    # where
    # Action - The action to perform.  R=Replace, D=Delete
    # Search - The string to search for.  Can be a regular expression
    # Replace - The string to replace the search string with.  Can use regex groups
    # Position - Where to insert the new string.  The new string is placed BEFORE the string specified in position

    # Master IP Address resource
    xforms.append(('R', r'^(.*?)IGNORE_ERROR[\s]+add[\s]+secondary[\s]+ipaddress[\s]+(.+?)$', r'\1 ignore_error add_secondary_ipaddress \2 -np', ''))
    xforms.append(('R', r'^(.*?)IGNORE_ERROR[\s]+del[\s]+secondary[\s]+ipaddress[\s]+(.+?)$', r'\1 ignore_error del_secondary_ipaddress \2', ''))
    xforms.append(('D', r'IGNORE_ERROR[\s]+set[\s]+allow[\s]+ip[\s]+address[\s]+duplicates', '', ''))
    xforms.append(('D', r'IGNORE_ERROR[\s]+NUDP', '', ''))
    xforms.append(('D', r'IGNORE_ERROR[\s]+HTTP', '', ''))

    # Everything else
    xforms.append(('R', r'^(.*?)add[\s]+secondary[\s]+ipaddress[\s]+(\S+)(.*)', r'CIFS_IP=\2\n\1exit_on_error add_secondary_ipaddress \2\3', r'nss /poolact'))
    xforms.append(('R', r'^(.*?)nss[\s]+/poolactivate[\s]*=[\s]*(.+?)$', r'\1exit_on_error nss /poolact=\2', ''))
    xforms.append(('R', r'^(.*?)mount[\s]+(.+?)[\s]+VOLID[\s]*=[\s]*(.+?)$', r'\1exit_on_error ncpcon mount \2=\3', ''))
    xforms.append(('R', r'^(.*?)NUDP[\s]+ADD[\s]+(.+?)[\s]+(.+?)$', r'\1exit_on_error ncpcon bind --ncpservername=\2 --ipaddress=\3', ''))
    xforms.append(('R', r'^(.*?)CIFS[\s]+ADD[\s]+(.+?)$', 'exit_on_error novcifs --add \'--vserver="\\2"\' --ip-addr=$CIFS_IP', ''))
    xforms.append(('R', r'^(.*?)AFPBIND[\s]+ADD[\s]+(.+?)$', r'exit_on_error cluster_afp.sh add \2', ''))

    xforms.append(('R', r'^(.*?)del[\s]+secondary[\s]+ipaddress[\s]+(\S+)(.*)', r'CIFS_IP=\2\n\1ignore_error del_secondary_ipaddress \2\3', ''))
    xforms.append(('R', r'^(.*?)CIFS[\s]+DEL[\s]+(.+?)$', 'ignore_error novcifs --remove \'--vserver="\\2"\' --ip-addr=$CIFS_IP', 'del_secondary_ipaddress'))
    xforms.append(('R', r'^(.*?)AFPBIND[\s]+DEL[\s]+(.+?)$', r'ignore_error cluster_afp.sh del \2', 'del_secondary_ipaddress'))
    xforms.append(('R', r'^(.*?)NUDP[\s]+DEL[\s]+(.+?)[\s]+(.+?)$', r'\1ignore_error ncpcon unbind --ncpservername=\2 --ipaddress=\3', 'del_secondary_ipaddress'))
    xforms.append(('R', r'^(.*?)nss[\s]+/pooldeactivate[\s]*=[\s]*(.+?)$', r'\1ignore_error nss /pooldeact=\2', 'del_secondary_ipaddress'))
    xforms.append(('D', r'CLUSTER CVSBIND', '', ''))

    try:
        fd = open( syntax_file, "r" )
        xform = list()
        for line in string.split( fd.read(), '\n' ):
            xform.append( line )
            if 4 == len( xform ):
                xforms.append((xform[0], xform[1], xform[2], xform[3]))
                xform = list()
        fd.close()
    except:
        pass

    # Look at each line to see if it should be modified
    oldScript = string.split(script, '\n')
    for line in oldScript:
        line = string.strip(line)
        if len(line) == 0:
            continue

        for xform in xforms:
            mo = re.search(xform[1], line, re.IGNORECASE)
            if mo:
                if xform[0] == 'D':
                    pass
                elif xform[0] == 'R':
                    str = mo.expand(xform[2])
                    if xform[3]:
                        # A position string was specified.  Insert before it.
                        for idx in range(len(newScript)):
                            mo = re.search(xform[3], newScript[idx])
                            if mo:
                                newScript.insert(idx, str + '\n')
                                break
                        else:
                            # Could not find the position string
                            newScript.append(str + '\n')
                    else:
                        # No position string was specified
                        ins_lines = string.split(str, '\n')
                        for item in ins_lines:
                            newScript.append(item + '\n')
                changed = True
                break
        else:

            # No RegEx string matched this line.  Pass it through
            newScript.append(line + '\n')

    if changed:
        # Determine if the header needs to be written
        for line in newScript:
            if line == '#!/bin/bash':
                break
        else:
            newScript.insert(0, '. /opt/novell/ncs/lib/ncsfuncs\n')
            newScript.insert(0, '#!/bin/bash\n')
            newScript.append('exit 0\n')

        script = ""
        for line in newScript:
            script += line

    return(changed, script)

#############################################################################
#
# Function: createVirtualNCPServer
#   Purpose:
#
def createVirtualNCPServer(ld, clusterDn, containerDn, serverName, ipAddress):

    ncpServerDn = "cn=" + serverName + "," + containerDn

    ip = string.split(ipAddress, ".")
    ipBytes = struct.pack("BBBB", int(ip[0]), int(ip[1]), int(ip[2]), int(ip[3]))

    attrs = []
    attrs.append( ("cn", serverName) )
    attrs.append( ("objectClass", "ncpServer") )
    attrs.append( ("version", "NWCS") )
    attrs.append( ("nCSNetWareCluster", clusterDn) )
    attrs.append( ("networkAddress", [ "8#\x02\x0c" + ipBytes, "9#\x02\x0c" + ipBytes ]) )

    ld.add_s(ncpServerDn, attrs)

    return ncpServerDn

#############################################################################
#
# Function: linkVolumeResourceAndVirtualNCPServer
#   Purpose:
#
def linkVolumeResourceAndVirtualNCPServer(ld, resourceDn, ncpServerDn):

    attrs = []
    attrs.append( (ldap.MOD_ADD, "Resource", resourceDn) )

    ld.modify_s(ncpServerDn, attrs)

    attrs = []
    attrs.append( (ldap.MOD_ADD, "nCSNCPServer", ncpServerDn) )

    ld.modify_s(resourceDn, attrs)

#############################################################################
#
# Function: linkPoolToVirtualNCPServer
#   Purpose:
#
def linkPoolToVirtualNCPServer(ld, poolDn, ncpServerDn):

    attrs = []
    attrs.append( (ldap.MOD_REPLACE, "hostServer", ncpServerDn) )

    ld.modify_s(poolDn, attrs)

#############################################################################
#
# Function: createOcfsConfFile
#
#   Purpose:
#
def createOcfsConfFile(ld, clusterDn):

    nodeCount = 0
    clusterName = objectNameFromLdapDn(clusterDn)

    res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer", [ "cn", "nCSGIPCNodeNumber", "nCSNetworkAddress" ] )

    os.system( "mkdir -p /etc/ocfs2/" )
    conf = file("/etc/ocfs2/cluster.conf", "w+")

    for node in res:
        if not (node[1].has_key( "nCSGIPCNodeNumber" ) and node[1].has_key( "nCSNetworkAddress" )):
            continue;
        nodeNumber = node[1]["nCSGIPCNodeNumber"][0]
        ipBytes = struct.unpack("BBBB", node[1]["nCSNetworkAddress"][0][2:])
        ipAddress = str(ipBytes[0]) + "." + str(ipBytes[1]) + "." + str(ipBytes[2]) + "." + str(ipBytes[3])

        nodeName = node[1]["cn"][0]

        conf.write("node:\n")
        conf.write("\tip_port = 7777\n")
        conf.write("\tip_address = " + ipAddress + "\n")
        conf.write("\tnumber = " + nodeNumber + "\n")
        conf.write("\tname = " + nodeName  + "\n")
        conf.write("\tcluster = " + clusterName + "\n")
        conf.write("\n")
        nodeCount = nodeCount + 1

    conf.write("cluster:\n");
    conf.write("\tnode_count = " + str(nodeCount) + "\n")
    conf.write("\tname = " + clusterName + "\n")
    conf.write("\n")

    conf.close()

    o2cb_conf_file_name="/etc/sysconfig/o2cb"

    #attempt to edit the variables first. Doesn't really matter if the file is not there.
    os.system( "sed -i -e 's/^\\([[:space:]]*O2CB_ENABLED[[:space:]]*=[[:space:]]*\)\\<\\w\\+\\>/\\1true/' " + o2cb_conf_file_name )
    os.system( "sed -i -e 's/^\\([[:space:]]*O2CB_BOOTCLUSTER[[:space:]]*=[[:space:]]*\\)\\<[^[:space:]]\\+/\\1" + clusterName + "/' " + o2cb_conf_file_name )

    #Search for the two variables we care
    ret_enabled = os.system ( "grep -e '^[[:space:]]*O2CB_ENABLED[[:space:]]*=[[:space:]]*true\\>' " + o2cb_conf_file_name + " &> /dev/null" )
    ret_bootcluster = os.system ( "grep -e '^[[:space:]]*O2CB_BOOTCLUSTER[[:space:]]*=[[:space:]]*" + clusterName + "\\>' " + o2cb_conf_file_name + " &> /dev/null" )

    #file not there?
    if 512 == ret_enabled:
        conf2 = file(o2cb_conf_file_name, "w+")
        conf2.write("#\n")
        conf2.write("# This is a configuration file for automatic startup of the O2CB\n")
        conf2.write("# driver.  It was generated by /opt/novell/ncs/bin/ncs-configd\n")
        conf2.write("# It is normally gererated by running /etc/init.d/o2cb configure.\n")
        conf2.write("\n")
        conf2.write("# O2CB_ENABLED: 'true' means to load the drive on boot.\n")
        conf2.write("O2CB_ENABLED=true\n")
        conf2.write("\n")
        conf2.write("# O2CB_BOOTCLUSTER: name of cluster to start\n")
        conf2.write("O2CB_BOOTCLUSTER=" + clusterName + "\n")
        conf2.write("\n")
        conf2.close()
    else:
    #file exists but missing some statements
        if 256 == ret_enabled:
            os.system( "echo 'O2CB_ENABLED=true' >> " + o2cb_conf_file_name )
        if 256 == ret_bootcluster:
            os.system( "echo 'O2CB_BOOTCLUSTER=" + clusterName + "' >> " + o2cb_conf_file_name )

#############################################################################
#
# Function: createGipcConfFile
#
#   Purpose:
#
def createGipcConfFile(ld, clusterDn):

    res = ld.search_s(clusterDn, ldap.SCOPE_BASE)
    guid = ld.search_s(clusterDn, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn(clusterDn), ["GUID"])

    conf = file(os.path.join(clstrlibPath, "gipc.conf"), "w+")

    # For mixed clusters, use existing panning ID if present
    if res[0][1].has_key("nCSGIPCConfig"):
        nwGipcConfig = res[0][1]["nCSGIPCConfig"][0]
        left = nwGipcConfig.find("panning")
        right = nwGipcConfig[left:].find("\n")
        conf.write(nwGipcConfig[left:][:right].strip() + "\r\n")
    # otherwise, generate panning ID from cluster container GUID
    else:
        cid = struct.unpack("I", guid[0][1]["GUID"][0][0:4])
        conf.write("panning clusterid " + str(cid[0]) + "\r\n")

    conf.write("heartbeat rate_usecs " + str(int(res[0][1]["nCSGIPCHeartbeat"][0]) * 1000000) + "\r\n")
    conf.write("censustaker tolerance " + str(int(res[0][1]["nCSGIPCTolerance"][0]) * 1000000) + "\r\n")
    conf.write("sequencer master_watchdog " + str(int(res[0][1]["nCSGIPCMasterWatchdog"][0]) * 1000000) + "\r\n")
    conf.write("sequencer slave_watchdog " + str(int(res[0][1]["nCSGIPCSlaveWatchdog"][0]) * 1000000) + "\r\n")
    conf.write("sequencer retrans_max " + res[0][1]["nCSGIPCMaxRetransmits"][0] + "\r\n")
    conf.write("\r\n")

    res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer", ["nCSGIPCNodeNumber", "nCSNetworkAddress"])
    for node in res:
        if not (node[1].has_key( "nCSGIPCNodeNumber" ) and node[1].has_key( "nCSNetworkAddress" )):
            continue;
        nodeNumber = node[1]["nCSGIPCNodeNumber"][0]
        ipBytes = struct.unpack("BBBB", node[1]["nCSNetworkAddress"][0][2:])
        ipAddress = str(ipBytes[0]) + "." + str(ipBytes[1]) + "." + str(ipBytes[2]) + "." + str(ipBytes[3])
        conf.write("nodeid " + ipAddress + " " + nodeNumber + "\r\n")
    conf.write("\r\n")

    conf.close()

#############################################################################
#
# Function: createResourcePriorityConfFile
#
#   Purpose:
#
def createResourcePriorityConfFile(ld, clusterDn):

    res = ld.search_s(clusterDn, ldap.SCOPE_BASE)

    conf = file(os.path.join(clstrlibPath, "resource-priority.conf"), "w+")
    conf.write(res[0][1]["nCSCRMResourcePriority"][0])
    conf.close()

#############################################################################
#
# Function: createClusterXmlFile
#
#   Purpose:
#
def createClusterXmlFile(ld, clusterDn, args=None):

    res = ld.search_s(clusterDn, ldap.SCOPE_BASE)
    guid = ld.search_s(clusterDn, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn(clusterDn), ["GUID"])
    res[0][1]["GUID"] = guid[0][1]["GUID"]

    xml = file(os.path.join(clstrlibPath, "cluster.xml"), "w+")

    d = dsml.DSMLWriter(xml, ["nCSNetworkAddress", "networkAddress", "GUID", "objectGUID"], "", "  ")
    d.writeHeader()
    d.writeRecord(res[0][0], res[0][1])
    d.writeFooter()

    xml.close()

#############################################################################
#
# Function: createNodesXmlFile
#
#   Purpose:
#
def createNodesXmlFile(ld, clusterDn, args=None):

    res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer" )

    xml = file(os.path.join(clstrlibPath, "nodes.xml"), "w+")

    d = dsml.DSMLWriter(xml, ["nCSNetworkAddress", "GUID", "objectGUID"], "", "  ")
    d.writeHeader()
    for node in res:
        d.writeRecord(node[0], node[1])
    d.writeFooter()

    xml.close()

#############################################################################
#
# Function: createResourcesXmlFile
#
#   Purpose:
#
def createResourcesXmlFile(ld, clusterDn, args=None):

    res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "(|(objectClass=nCSClusterResource)(objectClass=nCSVolumeResource)(objectClass=nCSResourceTemplate))" )

    xml = file(os.path.join(clstrlibPath, "resources.xml"), "w+")

    d = dsml.DSMLWriter(xml, ["nCSCRMPreferredNodes", "objectGUID" ], "", "  ")
    d.writeHeader()
    for resource in res:
        ncsAttributes = {}
        if resource[1].has_key( "cn" ):
            ncsAttributes["cn"] = resource[1]["cn"]
        if resource[1].has_key( "nCSBCCState" ):
            ncsAttributes["nCSBCCState"] = resource[1]["nCSBCCState"]
        if resource[1].has_key( "nCSCRMFailbackMode" ):
            ncsAttributes["nCSCRMFailbackMode"] = resource[1]["nCSCRMFailbackMode"]
        if resource[1].has_key( "nCSCRMFailoverMode" ):
            ncsAttributes["nCSCRMFailoverMode"] = resource[1]["nCSCRMFailoverMode"]
        if resource[1].has_key( "nCSCRMIgnoreQuorum" ):
            ncsAttributes["nCSCRMIgnoreQuorum"] = resource[1]["nCSCRMIgnoreQuorum"]
        if resource[1].has_key( "nCSCRMLoadTimeout" ):
            ncsAttributes["nCSCRMLoadTimeout"] = resource[1]["nCSCRMLoadTimeout"]
        if resource[1].has_key( "nCSCRMMonitorAction" ):
            ncsAttributes["nCSCRMMonitorAction"] = resource[1]["nCSCRMMonitorAction"]
        if resource[1].has_key( "nCSCRMMonitorFailurePeriod" ):
            ncsAttributes["nCSCRMMonitorFailurePeriod"] = resource[1]["nCSCRMMonitorFailurePeriod"]
        if resource[1].has_key( "nCSCRMMonitorFrequency" ):
            ncsAttributes["nCSCRMMonitorFrequency"] = resource[1]["nCSCRMMonitorFrequency"]
        if resource[1].has_key( "nCSCRMMonitorMaxFailures" ):
            ncsAttributes["nCSCRMMonitorMaxFailures"] = resource[1]["nCSCRMMonitorMaxFailures"]
        if resource[1].has_key( "nCSCRMMonitorTimeout" ):
            ncsAttributes["nCSCRMMonitorTimeout"] = resource[1]["nCSCRMMonitorTimeout"]
        if resource[1].has_key( "nCSCRMPreferredNodes" ):
            ncsAttributes["nCSCRMPreferredNodes"] = resource[1]["nCSCRMPreferredNodes"]
        if resource[1].has_key( "nCSCRMUnloadTimeout" ):
            ncsAttributes["nCSCRMUnloadTimeout"] = resource[1]["nCSCRMUnloadTimeout"]
        if resource[1].has_key( "nCSNCPServer" ):
            ncsAttributes["nCSNCPServer"] = resource[1]["nCSNCPServer"]
        if resource[1].has_key( "nCSRevision" ):
            ncsAttributes["nCSRevision"] = resource[1]["nCSRevision"]
        if resource[1].has_key( "objectClass" ):
            ncsAttributes["objectClass"] = resource[1]["objectClass"]
        d.writeRecord(resource[0], ncsAttributes )
    d.writeFooter()

    xml.close()

#############################################################################
#
# Function: updateResourcesXmlFile
#
#   Purpose:
#
def updateResourcesXmlFile(ld, clusterDn, resourceName):

    found = False

    resourceDn = "cn=" + resourceName + "," + clusterDn
    res = ld.search_s(resourceDn, ldap.SCOPE_BASE)

    tmpPath = "/tmp/" + resourceName + ".xml"
    tmp = file(tmpPath, "w+")

    d = dsml.DSMLWriter(tmp, ["nCSCRMPreferredNodes","objectGUID"], "", "  ")
    d.writeHeader()
    for resource in res:
        ncsAttributes = {}
        if resource[1].has_key( "cn" ):
            ncsAttributes["cn"] = resource[1]["cn"]
        if resource[1].has_key( "nCSBCCState" ):
            ncsAttributes["nCSBCCState"] = resource[1]["nCSBCCState"]
        if resource[1].has_key( "nCSBCCPeers" ):
            ncsAttributes["nCSBCCPeers"] = resource[1]["nCSBCCPeers"]
        if resource[1].has_key( "nCSBCCPriority" ):
            ncsAttributes["nCSBCCPriority"] = resource[1]["nCSBCCPriority"]
        if resource[1].has_key( "nCSBCCRevision" ):
            ncsAttributes["nCSBCCRevision"] = resource[1]["nCSBCCRevision"]
        if resource[1].has_key( "nCSCRMFailbackMode" ):
            ncsAttributes["nCSCRMFailbackMode"] = resource[1]["nCSCRMFailbackMode"]
        if resource[1].has_key( "nCSCRMFailoverMode" ):
            ncsAttributes["nCSCRMFailoverMode"] = resource[1]["nCSCRMFailoverMode"]
        if resource[1].has_key( "nCSCRMIgnoreQuorum" ):
            ncsAttributes["nCSCRMIgnoreQuorum"] = resource[1]["nCSCRMIgnoreQuorum"]
        if resource[1].has_key( "nCSCRMLoadTimeout" ):
            ncsAttributes["nCSCRMLoadTimeout"] = resource[1]["nCSCRMLoadTimeout"]
        if resource[1].has_key( "nCSCRMMonitorAction" ):
            ncsAttributes["nCSCRMMonitorAction"] = resource[1]["nCSCRMMonitorAction"]
        if resource[1].has_key( "nCSCRMMonitorFailurePeriod" ):
            ncsAttributes["nCSCRMMonitorFailurePeriod"] = resource[1]["nCSCRMMonitorFailurePeriod"]
        if resource[1].has_key( "nCSCRMMonitorFrequency" ):
            ncsAttributes["nCSCRMMonitorFrequency"] = resource[1]["nCSCRMMonitorFrequency"]
        if resource[1].has_key( "nCSCRMMonitorMaxFailures" ):
            ncsAttributes["nCSCRMMonitorMaxFailures"] = resource[1]["nCSCRMMonitorMaxFailures"]
        if resource[1].has_key( "nCSCRMMonitorTimeout" ):
            ncsAttributes["nCSCRMMonitorTimeout"] = resource[1]["nCSCRMMonitorTimeout"]
        if resource[1].has_key( "nCSCRMPreferredNodes" ):
            ncsAttributes["nCSCRMPreferredNodes"] = resource[1]["nCSCRMPreferredNodes"]
        if resource[1].has_key( "nCSCRMUnloadTimeout" ):
            ncsAttributes["nCSCRMUnloadTimeout"] = resource[1]["nCSCRMUnloadTimeout"]
        if resource[1].has_key( "nCSNCPServer" ):
            ncsAttributes["nCSNCPServer"] = resource[1]["nCSNCPServer"]
        if resource[1].has_key( "nCSRevision" ):
            ncsAttributes["nCSRevision"] = resource[1]["nCSRevision"]
        if resource[1].has_key( "nCSVolumes" ):
            ncsAttributes["nCSVolumes"] = resource[1]["nCSVolumes"]
        if resource[1].has_key( "objectClass" ):
            ncsAttributes["objectClass"] = resource[1]["objectClass"]
        d.writeRecord(resource[0], ncsAttributes )
    d.writeFooter()

    tmp.close()

    newDom = xml.dom.minidom.parse(tmpPath)
    newElement = newDom.getElementsByTagName("dsml:entry")[0]

    resourceCn = unicode( "cn=" + resourceName.lower() + ",", "utf-8" )
    for i in range( 1, 3 ):
        try:
            oldDom = xml.dom.minidom.parse(os.path.join(clstrlibPath, "resources.xml"))
            res = oldDom.getElementsByTagName("dsml:entry")
            break;
        except:
            createResourcesXmlFile( ld, clusterDn )

    for resource in res:
        if (resource.getAttribute("dn").lower().startswith(resourceCn)):
            resource.parentNode.replaceChild(newElement, resource)
            found = True
            break

    if (found == False):
        res[0].parentNode.appendChild(newElement)

    newFile = file(os.path.join(clstrlibPath, "resources.xml"), "w+")
    newFile.write(oldDom.toxml("utf-8") + "\n")
    newFile.close()
    os.remove(tmpPath)

#############################################################################
#
# Class: resourcesXmlHandler
#
#   Purpose:
#
class resourcesXmlHandler(xml.sax.handler.ContentHandler):

    def __init__(self):
        self.resName = None
        self.attrName = None
        self.attrValue = None
        self.flag = 0
        self.dsmlTag = None
        self.dsmlName = None
        self.dsmlValue = None

    def setParams(self, resName, attrName):
        self.resName = resName
        self.attrName = attrName

    def getResult(self):
        return self.attrValue

    def startElement(self, name, attrs):
        if (self.flag == 2):
            return
        self.dsmlTag = name
        if (self.dsmlTag == "dsml:attr"):
            self.dsmlName = attrs.getValue("name")

    def characters(self, ch):
        if (self.flag == 2):
            return
        if (self.dsmlTag != "dsml:value"):
            return
        el = string.strip(ch)
        if (len(el) == 0):
            return
        self.dsmlValue = el
        if (self.dsmlName == "cn"):
            if (self.dsmlValue == unicode( self.resName, "utf-8")):
                self.flag = 1
            else:
                self.flag = 0
        elif (self.dsmlName == self.attrName):
            if (self.flag == 1):
                self.flag = 2
                self.attrValue = self.dsmlValue

#############################################################################
#
# Function: getResourceAttribute
#
#   Purpose:
#
def getResourceAttribute(resourceName, attrName):

    value = None
    try:
        parser = xml.sax.make_parser()
        parser.setFeature(xml.sax.handler.feature_external_ges, False)
        handler = resourcesXmlHandler()
        handler.setParams(resourceName, attrName)
        parser.setContentHandler(handler)
        parser.parse(os.path.join(clstrlibPath, "resources.xml"))
        value = handler.getResult()
    except:
        pass

    if (value != None):
        return (True, value)
    else:
        return (False, value)

#############################################################################
#
# Function: getResourceRevision
#
#   Purpose:
#
def getResourceRevision(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSRevision")

    if found:
        return (True, int(value))
    else:
        return (False, 0)

#############################################################################
#
# Function: getResourceMonitorEnabled
#
#   Purpose:
#
def getResourceMonitorEnabled(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSCRMFailoverMode")

    if found:
        return bool(int(value) & 0x00000040)
    else:
        return False

#############################################################################
#
# Function: getResourceMonitorAction
#
#   Purpose:
#
def getResourceMonitorFailover(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSCRMFailoverMode")

    if found:
        if (int(value) & 0x00000080):
            return 1
        elif (int(value) & 0x00000200):
            return 2

    return 0

#############################################################################
#
# Function: getResourceMonitorFrequency
#
#   Purpose:
#
def getResourceMonitorFrequency(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSCRMMonitorFrequency")

    if found:
        return int(value)
    else:
        return 600


#############################################################################
#
# Function: getResourceMonitorTimeout
#
#   Purpose:
#
def getResourceMonitorTimeout(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSCRMMonitorTimeout")

    if found:
        return int(value)
    else:
        return 600

#############################################################################
#
# Function: getResourceScriptTimeout
#
#   Purpose:
#
def getResourceScriptTimeout(resourceName, load_unload ):
    if (load_unload == "load"):
        found, value = getResourceAttribute(resourceName, "nCSCRMLoadTimeout")
    else:
        found, value = getResourceAttribute(resourceName, "nCSCRMUnloadTimeout")

    if found:
        return int(value)
    else:
        return 60

#############################################################################
#
# Function: getResourceMonitorMaxFailures
#
#   Purpose:
#
def getResourceMonitorMaxFailures(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSCRMMonitorMaxFailures")

    if found:
        return int(value)
    else:
        return 1

#############################################################################
#
# Function: getResourceMonitorFailurePeriod
#
#   Purpose:
#
def getResourceMonitorFailurePeriod(resourceName):
    found, value = getResourceAttribute(resourceName, "nCSCRMMonitorFailurePeriod")

    if found:
        return int(value)
    else:
        return 600

#############################################################################
#
# Function: haveResourceScripts
#
#   Purpose:
#
def haveResourceScripts(resourceName):

    try:
        os.stat(os.path.join(clstrlibPath, resourceName + ".load"))
        os.stat(os.path.join(clstrlibPath, resourceName + ".unload"))
        os.stat(os.path.join(clstrlibPath, resourceName + ".monitor"))
        return True
    except:
        return False

#############################################################################
#
# Function: createResourceScripts
#
#   Purpose:
#
def createResourceScripts(ld, clusterDn, resourceName=None):

    staleResourceFileNames = []

    # Bug 302258
    # Need to change the CWD to the same directory as the resource
    # scripts.  dos2unix creates a temp file in the CWD of the process
    # and then renames the temp file to the destination.  The rename
    # will fail if the CWD is on a different device than the resource
    # script directory.
    oldPath = os.getcwd()
    os.chdir(clstrlibPath)

    if (resourceName == None):

        res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "(|(objectClass=nCSClusterResource)(objectClass=nCSVolumeResource)(objectClass=nCSResourceTemplate))" )

        for fileName in os.listdir(clstrlibPath):

            if fileName.endswith(".load"):
                staleResourceFileNames.append(fileName)

            if fileName.endswith(".unload"):
                staleResourceFileNames.append(fileName)

            if fileName.endswith(".monitor"):
                staleResourceFileNames.append(fileName)

    else:

        resourceDn = "cn=" + resourceName + "," + clusterDn
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)

    for resource in res:

        resourceName = resource[1]["cn"][0]

        if resource[1].has_key("nCSCRMLoadScript"):

            loadName = resourceName + ".load"
            try:
                staleResourceFileNames.remove(loadName)
            except:
                pass

            loadPath = os.path.join(clstrlibPath, loadName)
            load = file(loadPath, "w+")
            load.write(resource[1]["nCSCRMLoadScript"][0])
            load.close()
            os.system("dos2unix " + loadPath)
            os.chmod(loadPath, 0755)

        if resource[1].has_key("nCSCRMUnloadScript"):

            unloadName = resourceName + ".unload"
            try:
                staleResourceFileNames.remove(unloadName)
            except:
                pass

            unloadPath = os.path.join(clstrlibPath, unloadName)
            unload = file(unloadPath, "w+")
            unload.write(resource[1]["nCSCRMUnloadScript"][0])
            unload.close()
            os.system("dos2unix " + unloadPath)
            os.chmod(unloadPath, 0755)

        if resource[1].has_key("nCSCRMMonitorScript"):

            monitorName = resourceName + ".monitor"
            try:
                staleResourceFileNames.remove(monitorName)
            except:
                pass

            monitorPath = os.path.join(clstrlibPath, monitorName)
            monitor = file(monitorPath, "w+")
            monitor.write(resource[1]["nCSCRMMonitorScript"][0])
            monitor.close()
            os.system("dos2unix " + monitorPath)
            os.chmod(monitorPath, 0755)

    for fileName in staleResourceFileNames:

        os.remove(os.path.join(clstrlibPath, fileName))

    os.chdir(oldPath)

#############################################################################
#
# Function: updateClusterConfig
#
#   Purpose:
#
def updateClusterConfig(ld, clusterDn):

    try:
        createGipcConfFile(ld, clusterDn)
    except:
        pass
    try:
        createOcfsConfFile(ld, clusterDn)
    except:
        pass
    try:
        createResourcePriorityConfFile(ld, clusterDn)
    except:
        pass
    try:
        createClusterXmlFile(ld, clusterDn)
    except:
        pass
    try:
        createNodesXmlFile(ld, clusterDn)
    except:
        pass
    try:
        createResourcesXmlFile(ld, clusterDn)
    except:
        pass
    try:
        createResourceScripts(ld, clusterDn)
    except:
        pass

#############################################################################
#
# Function: updateResourceConfig
#
#   Purpose:
#
def updateResourceConfig(ld, clusterDn, resourceName):

    try:
        updateResourcesXmlFile(ld, clusterDn, resourceName)
    except:
        pass
    try:
        createResourceScripts(ld, clusterDn, resourceName)
    except:
        pass
#############################################################################
#
# Function: clusterEnableSteps
#
#   Purpose:
#
def clusterEnableSteps(ld, clusterDn, poolDn, resourceName, poolName, serverName, ipAddress, advertize):

    retval = 0
    containerDn = containerFromLdapDn(clusterDn)
    resourceDn = ""
    ncpServerDn = ""

    try:
        resourceDn = createVolumeResource(ld, clusterDn, resourceName, poolName, serverName, ipAddress, int( advertize ))

        try:
            ncpServerDn = createVirtualNCPServer(ld, clusterDn, containerDn, serverName, ipAddress)

            try:
                linkVolumeResourceAndVirtualNCPServer(ld, resourceDn, ncpServerDn)

                try:
                    linkPoolToVirtualNCPServer(ld, poolDn, ncpServerDn)

                except Exception, e:
                    retval = -4
                    syslog.syslog(syslog.LOG_DEBUG, "linkPoolToVirtualNCPServer failed: %s" % (str(e)))

            except Exception, e:
                retval = -3
                syslog.syslog(syslog.LOG_DEBUG, "linkVolumeResourceAndVirtualNCPServer failed: %s" % (str(e)))

        except Exception, e:
            retval = -2
            syslog.syslog(syslog.LOG_DEBUG, "createVirtualNCPServer failed: %s" % (str(e)))

    except Exception, e:
        retval = -1
        syslog.syslog(syslog.LOG_DEBUG, "createVolumeResource failed: %s" % (str(e)))

    if (retval != 0):
        if (ncpServerDn != ""):
            try:
                ld.delete_s(ncpServerDn)
            except:
                pass

        if (resourceDn != ""):
            try:
                ld.delete_s(resourceDn)
            except:
                pass

    return retval


#############################################################################
#
# Function: clusterEnableSteps2
#
#   Purpose: same as clsuterEnableSteps but with CIFS changes.
#
def clusterEnableSteps2(ld, clusterDn, poolDn, resourceName, poolName, serverName, cifsServerName, ipAddress, advertize):

    retval = 0
    containerDn = containerFromLdapDn(clusterDn)
    resourceDn = ""
    ncpServerDn = ""

    try:
        resourceDn = createVolumeResource(ld, clusterDn, resourceName, poolName, serverName, ipAddress, int( advertize ))

        try:
            ncpServerDn = createVirtualNCPServer(ld, clusterDn, containerDn, serverName, ipAddress)

            try:
                linkVolumeResourceAndVirtualNCPServer(ld, resourceDn, ncpServerDn)

                try:
                    linkPoolToVirtualNCPServer(ld, poolDn, ncpServerDn)

                    if 2 & int( advertize ) : # CIFS ADVERTIZE
                        try :
                            cifsEnableSteps(ld, ncpServerDn, ipAddress, cifsServerName)
                        except Exception, e:
                            retval =-5
                            syslog.syslog(syslog.LOG_DEBUG, "cifsEnableSteps failed: %s" % (str(e)))
                    else:
                        try :
                            cifsServerName = objectNameFromLdapDn( clusterDn ) + "-" + poolName + "-W"
                            start = len( cifsServerName )
                            if ( start > 15 ):
                                start -= 15
                            else:
                                start = 0

                            if( '.' == cifsServerName[start] or '-' == cifsServerName[start] or '_' == cifsServerName[start] ):
                                if( 0 == start ):
                                    num = 1;
                                    while( '.' == cifsServerName[num] or '-' == cifsServerName[num] or '_' == cifsServerName[num] ):
                                        num += 1

                                    cifsServerName = cifsServerName[num] + cifsServerName[1:]

                                else:
                                    num = start-1
                                    while( '.' == cifsServerName[num] or '-' == cifsServerName[num] or '_' == cifsServerName[num] ):
                                        num -= 1

                                    cifsServerName = cifsServerName[:start] + cifsServerName[num] + cifsServerName[start+1:]

                            cifsEnableSteps(ld, ncpServerDn, ipAddress, cifsServerName[start:].upper().replace( '_', '-' ))
                        except Exception, e:
                            pass

                except Exception, e:
                    retval = -4
                    syslog.syslog(syslog.LOG_DEBUG, "linkPoolToVirtualNCPServer failed: %s" % (str(e)))

            except Exception, e:
                retval = -3
                syslog.syslog(syslog.LOG_DEBUG, "linkVolumeResourceAndVirtualNCPServer failed: %s" % (str(e)))

        except Exception, e:
            retval = -2
            syslog.syslog(syslog.LOG_DEBUG, "createVirtualNCPServer failed: %s" % (str(e)))

    except Exception, e:
        retval = -1
        syslog.syslog(syslog.LOG_DEBUG, "createVolumeResource failed: %s" % (str(e)))

    if (retval != 0):
        if (ncpServerDn != ""):
            try:
                ld.delete_s(ncpServerDn)
            except:
                pass

        if (resourceDn != ""):
            try:
                ld.delete_s(resourceDn)
            except:
                pass

    return retval

#############################################################################
#
# Function: cifsEnableSteps
#
#   Purpose:
#
def cifsEnableSteps(ld, ncpServerDn, ipAddress, cifsServerName):

    retval = 0
    ip = string.split(ipAddress, ".")
    ipBytes = struct.pack("BBBB", int(ip[0]), int(ip[1]), int(ip[2]), int(ip[3]))
    res = ld.search_s( ncpServerDn, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn( ncpServerDn ), [ "objectClass", "nfapCIFSAttach", "nfapCIFSComment", "nfapCIFSServerName", "ACL" ] )

    attrs = []
    if not res[0][1].has_key( "nfapCIFSServerName" ):
        attrs.append( (ldap.MOD_ADD,"nfapCIFSServerName", cifsServerName ))

    if res[0][1].has_key( "nfapCIFSAttach" ):
        try:
            res[0][1]["nfapCIFSAttach"].index( "1#" + ipBytes )
        except:
            res[0][1]["nfapCIFSAttach"].append(  "1#" + ipBytes )
            attrs.append( (ldap.MOD_REPLACE, "nfapCIFSAttach",  res[0][1]["nfapCIFSAttach"] ))
    else:
        attrs.append( (ldap.MOD_ADD, "nfapCIFSAttach", [ "1#" + ipBytes ]))

    if res[0][1].has_key( "nfapCIFSComment" ):
        attrs.append( (ldap.MOD_REPLACE, "nfapCIFSComment", "Cluster Virtual CIFS Server" ))
    else:
        attrs.append( (ldap.MOD_ADD,"nfapCIFSComment","Cluster Virtual CIFS Server" ) )

    try:
        res[0][1]["objectClass"].index( "nfapCIFSConfigInfo" )
    except:
        attrs.append( ( ldap.MOD_ADD,"objectClass", "nfapCIFSConfigInfo") )

    if res[0][1].has_key( "ACL" ):
        try:
            res[0][1]["ACL"].index( '3#entry#[Root]#nfapCIFSServerName' )
        except:
            res[0][1]["ACL"].append( '3#entry#[Root]#nfapCIFSServerName' )
            attrs.append( (ldap.MOD_REPLACE, "ACL", res[0][1]["ACL"]) )
    else:
        attrs.append( (ldap.MOD_ADD, "ACL", "3#entry#[Root]#nfapCIFSServerName") )

    ld.modify_s(ncpServerDn, attrs)

#############################################################################
#
# Function: NWCLSTR_UpdateCIFSAttributes
#
#   Purpose:
#
def NWCLSTR_UpdateCIFSAttributes(ld, clusterDn, args):
    cifsServerName = objectNameFromLdapDn( clusterDn ) + "-" + NWCLSTR_GetRawPoolName( ld, clusterDn, args ) + "-W"
    start = len( cifsServerName )
    if ( start > 15 ):
        start -= 15
    else:
        start = 0

    if( '.' == cifsServerName[start] or '-' == cifsServerName[start] or '_' == cifsServerName[start] ):
        if( 0 == start ):
            num = 1;
            while( '.' == cifsServerName[num] or '-' == cifsServerName[num] or '_' == cifsServerName[num] ):
                num += 1

            cifsServerName = cifsServerName[num] + cifsServerName[1:]

        else:
            num = start-1
            while( '.' == cifsServerName[num] or '-' == cifsServerName[num] or '_' == cifsServerName[num] ):
                num -= 1

            cifsServerName = cifsServerName[:start] + cifsServerName[num] + cifsServerName[start+1:]

    cifsEnableSteps( ld, args[2], args[3], cifsServerName[start:].upper().replace( '_', '-' ))

#############################################################################
#
# Function: addResource
#
#   Purpose:
#
def addResource(resourceName):

    xml = "<virtualIO><datastream name=\"command\"/></virtualIO>"
    xml += "<ncsRequest><clusterConfig><bccAddResource>"
    xml += "<name>" + resourceName + "</name><revision>1</revision>"
    xml += "</bccAddResource></clusterConfig></ncsRequest>"

    # fd = os.open("/tmp/admin", os.O_RDWR)
    fd = os.open("/admin/Novell/Cluster/ClusterConfig.cmd", os.O_RDWR)
    os.write(fd, xml)
    os.lseek(fd, 0, 0)
    results = os.read(fd, 512)
    os.close(fd)
    # print results

#############################################################################
#
# Function: removeResource
#
#   Purpose:
#
def removeResource(resourceName):

    xml = "<virtualIO><datastream name=\"command\"/></virtualIO>"
    xml += "<ncsRequest><clusterConfig><bccRemoveResource>"
    xml += "<name>" + resourceName + "</name>"
    xml += "</bccRemoveResource></clusterConfig></ncsRequest>"

    # fd = os.open("/tmp/admin", os.O_RDWR)
    fd = os.open("/admin/Novell/Cluster/ClusterConfig.cmd", os.O_RDWR)
    os.write(fd, xml)
    os.lseek(fd, 0, 0)
    results = os.read(fd, 512)
    os.close(fd)
    # print results

#############################################################################
#
# Function: updateResource
#
#   Purpose:
#
def updateResource(resourceName, revision):

    xml = "<virtualIO><datastream name=\"command\"/></virtualIO>"
    xml += "<ncsRequest><clusterConfig><updateClusterResource>"
    xml += "<name>" + resourceName + "</name><revision>" + str(revision) + "</revision>"
    xml += "</updateClusterResource></clusterConfig></ncsRequest>"

    # fd = os.open("/tmp/admin", os.O_RDWR)
    fd = os.open("/admin/Novell/Cluster/ClusterConfig.cmd", os.O_RDWR)
    os.write(fd, xml)
    os.lseek(fd, 0, 0)
    results = os.read(fd, 512)
    os.close(fd)
    # print results


#############################################################################
#
# Function: syncObject
#
#   Purpose:
#
def syncObject(ld, objectDn):

    # BUG 510386 :

    # 1) Need to take care of NDAP / LDAP naming convention
    objectDn = objectDn.replace('s=','st=')

    # 2) Need to give eDir more time to sync
    #    Changing retry count from 5 to 20
    retry = 20

    while (retry > 0):
        try:
            res = ld.search_s(objectDn, ldap.SCOPE_BASE)
            return res[0][0]
        except:
            pass
        time.sleep(1)
        retry = retry - 1
        syslog.syslog( syslog.LOG_DEBUG, "syncObject(" + objectDn + ") not found, retry=" + str( retry ) + ".\n" )

    return False

#############################################################################
#
# Function: syncResource
#
#   Purpose:
#
def syncResource(resourceName, state, exception = [False]):

    retry = 5
    exception[0] = False
    while (retry > 0):
        try:
            results = subprocess.Popen( [ "cluster", "status", resourceName ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
            if (results.upper().find(state.upper()) != -1):
                return True
            syslog.syslog( syslog.LOG_DEBUG, "syncResource(" + resourceName + ") not in state '" + state + "', retry=" + str( retry ) + ".\n" )
        except:
            syslog.syslog( syslog.LOG_ERROR, " Exception occured in syncResource retry=" + str(retry) + "\n")
            if (retry == 1):
                exception[0] = True
        time.sleep(2)
        retry = retry - 1

    return False

#############################################################################
#
# NWCLSTR_* upcalls invoked from the kernel clstrlib.ko thru ncs-configd
#
#############################################################################

#############################################################################
#
# Function: NWCLSTR_CRMNewRes
#
#   Purpose:
#
def NWCLSTR_CRMNewRes(ld, clusterDn, args):

    resourceName = args[1]
    resourceDn = "cn=" + resourceName + "," + clusterDn
    # Bug 510386 -
    # Need to take care of ldap/ndap conventions
    resourceDn = resourceDn.replace('s=','st=')


    updateResourceConfig(ld, clusterDn, resourceName)

    try:
        if (resourceDn == ld.search_s(resourceDn, ldap.SCOPE_BASE)[0][0]):
            return 0

    except:
        return -1

    return -2

#############################################################################
#
# Function: NWCLSTR_IsBCCLoaded
#
#   Purpose:
#
def NWCLSTR_FileExists(ld, clusterDn, args):

    filename = args[1]
    retval = 0
    try:
        if (os.path.isfile( filename )):
            retval = 1

    except:
        pass

    return retval

#############################################################################
#
# Function: NWCLSTR_UpdateRes
#
#   Purpose:
#   Notes:  NewFlag
#           0 edit existing resource
#           1 create regular resource,
#           2 create volume resource,
#           3 create resource template.
#
def NWCLSTR_UpdateRes(ld, clusterDn, args):

    if(NWCLSTR_SafeToUpgradeFS("OES2018SP3") == 0):
        EDIR_ATTR_LIMIT = CLSTRCONF_SCRIPTS_BUFFER_LEN
    else:
        EDIR_ATTR_LIMIT = EDIR_ATTR_LIMIT_OLD

    resourceName = args[1]
    resourceDn = "cn=" + resourceName + "," + clusterDn

    try:
        path = os.path.join(clstrlibPath, resourceName)

        f = file(path + ".load", "r")
        loadScript = f.read()
        f.close()

        f = file(path + ".unload", "r")
        unloadScript = f.read()
        f.close()

    except Exception, e:
        syslog.syslog(syslog.LOG_DEBUG, "Error reading script files: %s" % (str(e)))
        return -1

    try:
        monitorScript = "exit 0\n"
        f = file(path + ".monitor", "r")
        monitorScript = f.read()
        f.close()
    except Exception, e:
        syslog.syslog(syslog.LOG_DEBUG, "Error reading script files: %s" % (str(e)))
        pass


    newFlag = int(args[2])
    objectClass = [ "", "nCSClusterResource", "nCSVolumeResource", "nCSResourceTemplate" ][newFlag]

    if (newFlag > 0):
        try:
            attrs = []
            attrs.append( ("cn", resourceName) )
            attrs.append( ("objectClass", objectClass) )
            attrs.append( ("nCSCRMLoadTimeout", args[3]) )
            attrs.append( ("nCSCRMLoadScript", loadScript[:EDIR_ATTR_LIMIT]) )
            attrs.append( ("nCSCRMUnloadTimeout", args[4]) )
            attrs.append( ("nCSCRMUnloadScript", unloadScript[:EDIR_ATTR_LIMIT]) )
            attrs.append( ("nCSCRMMonitorTimeout", args[5]) )
            attrs.append( ("nCSCRMMonitorScript", monitorScript[:EDIR_ATTR_LIMIT]) )
            attrs.append( ("nCSCRMFailbackMode", "0") )
            attrs.append( ("nCSCRMFailoverMode", args[6]) )
            attrs.append( ("nCSCRMIgnoreQuorum", args[7]) )
            attrs.append( ("nCSCRMMonitorFrequency", args[8]) )
            attrs.append( ("nCSCRMMonitorMaxFailures", args[9]) )
            attrs.append( ("nCSCRMMonitorFailurePeriod", args[10]) )
            attrs.append( ("nCSRevision", args[11]) )
            attrs.append( ("nCSCRMPreferredNodes", args[12]) )

            ld.add_s(resourceDn, attrs)

        except Exception, e:
            syslog.syslog(syslog.LOG_DEBUG, "ldap.add_s: %s" % (str(e)))
            return -2

    else:
        try:
            attrs = []
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadTimeout", args[3]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadScript", loadScript[:EDIR_ATTR_LIMIT]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMUnloadTimeout", args[4]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMUnloadScript", unloadScript[:EDIR_ATTR_LIMIT]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorTimeout", args[5]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorScript", monitorScript[:EDIR_ATTR_LIMIT]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMFailoverMode", args[6]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMIgnoreQuorum", args[7]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorFrequency", args[8]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorMaxFailures", args[9]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorFailurePeriod", args[10]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSRevision", args[11]) )
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMPreferredNodes", args[12]) )

            ld.modify_s(resourceDn, attrs)

        except Exception, e:
            syslog.syslog(syslog.LOG_DEBUG, "ldap.modify_s: %s" % (str(e)))
            return -3

    try:
        updateResourceConfig(ld, clusterDn, resourceName)

    except:
        return -4

    return 0

#############################################################################
#
# Function: NWCLSTR_SetOnOffLine
#
#   Purpose:
#
def NWCLSTR_SetOnOffLine(ld, clusterDn, args):

    resourceName = args[1]
    resourceDn = "cn=" + resourceName + "," + clusterDn
    onOff = args[2]

    try:
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        mode = int(res[0][1]["nCSCRMFailoverMode"][0])

    except:
        return -1

    if (onOff == "1"):
        newMode = mode | 1
    else:
        newMode = mode & ~1

    try:
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMFailoverMode", str(newMode)) )

        ld.modify_s(resourceDn, attrs)

    except:
        return -2

    try:
        updateResourceConfig(ld, clusterDn, resourceName)

    except:
        return -3

    return 0

#############################################################################
#
# Function: NWCLSTR_NewConfig
#
#   Purpose:
#
def NWCLSTR_NewConfig(ld, clusterDn, args):

    writeFlag = int(args[1])

    if (writeFlag == 0):

        try:
            createResourcePriorityConfFile(ld, clusterDn)

        except:
            return -1

    else:

        try:
            path = os.path.join(clstrlibPath, "resource-priority.conf")

            f = file(path, "r")
            conf = f.read()
            f.close()

        except:
            return -2

        try:
            attrs = []
            attrs.append( (ldap.MOD_REPLACE, "nCSCRMResourcePriority", conf) )

            ld.modify_s(clusterDn, attrs)

        except Exception, e:

            syslog.syslog(syslog.LOG_INFO, "nCSCRMResourcePriority update failed")
            syslog.syslog(syslog.LOG_INFO, str(e))
            return -3

    return 0

#############################################################################
#
# Function: NWCLSTR_DelResource
#
#   Purpose:
#
def NWCLSTR_DelResource(ld, clusterDn, args):

    resourceDn = "cn=" + args[1] + "," + clusterDn

    try:
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        for volume in res[0][1]["nCSVolumes"]:
            try:
                ld.delete_s(volume)
            except:
                pass

    except:
        pass

    try:
        res = ld.delete_s(resourceDn)

    except:
        return -1

    try:
        createResourcesXmlFile(ld, clusterDn)

    except:
        return -2

    try:
        pid = os.fork()
        if( 0 == pid ):
            daemonize()
            os._exit( os.system( "echo -n \"exec rm -f " + os.path.join(clstrlibPath, args[1] ) + ".*\"" + " > /proc/ncs/cluster" ))
        else:
            os.waitpid( pid, 0 )
    except:
        return -3

    return 0

#############################################################################
#
# Function: NWCLSTR_DelObject
#
#   Purpose:
#
def NWCLSTR_DelObject(ld, clusterDn, args):

    try:
        res = ld.delete_s(eDirToLdapDn(args[1]))

    except:
        return -1

    return 0

#############################################################################
#
# Function: NWCLSTR_EmailFilter
#
#   Purpose:
#
def NWCLSTR_EmailFilter(ld, clusterDn, args):

    emailFilter = args[1]

    try:
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSEmailFilter", emailFilter) )

        ld.modify_s(clusterDn, attrs)

    except:
        return -1

    try:
        createClusterXmlFile(ld, clusterDn)

    except:
        return -2

    return 0

#############################################################################
#
# Function: NWCLSTR_EmailAddresses
#
#   Purpose:
#
def NWCLSTR_EmailAddresses(ld, clusterDn, args):

    try:
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSEmailAddresses", args[1:]) )

        ld.modify_s(clusterDn, attrs)

    except:
        return -1

    try:
        createClusterXmlFile(ld, clusterDn)

    except:
        return -2

    return 0

#############################################################################
#
# Function: NWCLSTR_ModifyVersion
#
#   Purpose:
#
def NWCLSTR_ModifyVersion(ld, clusterDn, args):

    try:
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSRevision", args[1:]) )

        ld.modify_s(clusterDn, attrs)

    except:
        return -1

    try:
        createClusterXmlFile(ld, clusterDn)

    except:
        return -2

    return 0

#############################################################################
#
# Function: NWCLSTR_UpgradePartitions
#
#   Purpose:
#
def NWCLSTR_UpgradePartitions(ld, clusterDn, args):
    try:
        os.system( "/sbin/.upg60part" )

    except:
        return -1

    return 0

#############################################################################
#
# Function: NWCLSTR_UpdateConfig
#
#   Purpose:
#
def NWCLSTR_UpdateConfig(ld, clusterDn, args):

    try:
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSGIPCTolerance", args[1]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSGIPCHeartbeat", args[2]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMQuorum", args[3]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMQuorumTimeout", args[4]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSGIPCMasterWatchdog", args[5]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSGIPCSlaveWatchdog", args[6]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSGIPCMaxRetransmits", args[7]) )

        ld.modify_s(clusterDn, attrs)

    except:
        return -1

    try:
        createClusterXmlFile(ld, clusterDn)

    except:
        return -2

    try:
        createGipcConfFile(ld, clusterDn)

    except:
        return -3

    return 0

#############################################################################
#
# Function: NWCLSTR_ClusterIPAddress
#
#   Purpose:
#
def NWCLSTR_ClusterIPAddress(ld, clusterDn, args):

    ipAddress = args[1]
    portNumber = args[2]

    ip = string.split(ipAddress, ".")
    ipBytes = struct.pack("BBBB", int(ip[0]), int(ip[1]), int(ip[2]), int(ip[3]))

    try:
        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSNetworkAddress", "1#" + ipBytes) )
        attrs.append( (ldap.MOD_REPLACE, "networkAddress", [ "8#\x02\x0c" + ipBytes, "9#\x02\x0c" + ipBytes ]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSPortNumber", portNumber) )

        ld.modify_s(clusterDn, attrs)

    except:
        return -1

    try:
        createClusterXmlFile(ld, clusterDn)

    except:
        return -2

    return 0

#############################################################################
#
# Function: NWCLSTR_GetRawPoolName
#
#   Purpose:
#
def NWCLSTR_GetRawPoolName(ld, clusterDn, args):

    poolDn = eDirToLdapDn(args[1])
    if '' == poolDn:
        return -3

    try:
        for i in range( 1, 20 ):
            res = ld.search_s(poolDn, ldap.SCOPE_BASE)
            if res[0][1].has_key("hostResourceName"):
                return res[0][1]["hostResourceName"][0]

            syslog.syslog( syslog.LOG_DEBUG, "NWCLSTR_GetRawPoolName(" + poolDn + ") hostResourceName not found, retry=" + str( i ) + ".\n" )
            time.sleep( 1 )

    except:
        return -1

    return -2

#############################################################################
#
# Function: NWCLSTR_ResVolServ
#
#   Purpose:
#
def NWCLSTR_ResVolServ(ld, clusterDn, args):

    resourceName = args[1]
    resourceDn = "cn=" + resourceName + "," + clusterDn

    try:
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        if res[0][1].has_key("nCSNCPServer"):
            return res[0][1]["nCSNCPServer"][0]

    except:
        return -1

    return -2

#############################################################################
#
# Function: NWCLSTR_ResVolPool
#
#   Purpose:
#
def NWCLSTR_ResVolPool(ld, clusterDn, args):

    serverDn = args[1]
    search_domains = set()
    search_domains.add( containerFromLdapDn(serverDn))
    search_domains.add( containerFromLdapDn(clusterDn))

    try:
        res = ld.search_s( clusterDn, ldap.SCOPE_ONELEVEL, "objectClass=nCSNCPServer", [ "nCSNCPServer" ] )
        for node in res:
            if node[1].has_key( "nCSNCPServer" ):
                search_domains.add( containerFromLdapDn( node[1]["nCSNCPServer"][0] ))
    except:
        pass

    for domain in search_domains:
        try:
            res = ld.search_s(domain, ldap.SCOPE_ONELEVEL, "(&(objectClass=nssfsPool)(hostServer=" + serverDn + "))", [ "hostServer" ] )

            for pool in res:
                if pool[1].has_key("hostServer"):
                    if (serverDn == pool[1]["hostServer"][0]):
                        return pool[0]

        except:
            return -1

    return -2

#############################################################################
#
# Function: NWCLSTR_PortalVolServ
#
#   Purpose:
#
def NWCLSTR_PortalVolServ(ld, clusterDn, args):

    ncpServerDn = args[1]
    writeFlag = args[2]

    if (writeFlag == "1"):

        ipAddress = args[3]
        ip = string.split(ipAddress, ".")
        ipBytes = struct.pack("BBBB", int(ip[0]), int(ip[1]), int(ip[2]), int(ip[3]))

        try:
            attrs = []
            attrs.append( (ldap.MOD_REPLACE, "networkAddress", [ "8#\x02\x0c" + ipBytes, "9#\x02\x0c" + ipBytes ]) )

            res = ld.search_s( ncpServerDn, ldap.SCOPE_BASE, "cn=" + objectNameFromLdapDn( ncpServerDn ), [ "nfapCIFSAttach" ] )
            if res[0][1].has_key( "nfapCIFSAttach" ):
                attrs.append( (ldap.MOD_REPLACE, "nfapCIFSAttach",  [ "1#" + ipBytes ] ))
            else:
                attrs.append( (ldap.MOD_ADD, "nfapCIFSAttach", [ "1#" + ipBytes ]))

            ld.modify_s(ncpServerDn, attrs)

        except:
            return -1

    else:

        try:
            res = ld.search_s(ncpServerDn, ldap.SCOPE_BASE)
            if res[0][1].has_key("networkAddress"):
                for networkAddress in res[0][1]["networkAddress"]:
                    if (networkAddress[:2] == "8#"):
                        return str(struct.unpack("I", networkAddress[4:])[0])

        except:
            return -2

    return 0

#############################################################################
#
# Function: NWCLSTR_UsedIP
#
#   Purpose:
#
def NWCLSTR_UsedIP(ld, clusterDn, args):

    ipAddress = args[1]

    try:
        results = subprocess.Popen( ["/bin/sh", "-c", "grep -i \"\\<" + ipAddress + "\\>\" /var/opt/novell/ncs/*.load" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
    except:
        results = "";

    # IP address found in script?
    if (results.find(ipAddress) != -1):
        return 1

    else:
        return 0

#############################################################################
#
# Function: NWCLSTR_GetLoadScriptName
#
#   Purpose: will return Loadscript name containing poolname provided
#
def NWCLSTR_GetLoadScriptName(poolName):
    try:
	return subprocess.Popen( [ "/bin/sh", "-c", "grep -e \"^[^#].*nss[[:space:]]\\+\\/poolact[^\>]*\\>=\\<" + poolName + "\\>\" /var/opt/novell/ncs/*.load |cut -d':' -f1" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0];
    except:
        return "";


#############################################################################
#
# Function: NWCLSTR_GetPoolsFromLoadScript
#
#   Purpose: will return list of pools containing load script path provided
#	     make sure the path expected is absolute path
#
def NWCLSTR_GetPoolsFromLoadScript(loadScriptPath):
    try:
	results = subprocess.Popen( [ "/bin/sh", "-c", "awk '/poolact/ && !/#/' " + loadScriptPath   ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0];
    except:
        return "";
    return re.findall("poolact[\w]*=(\S+)",results);

#############################################################################
#
# Function: NWCLSTR_IsPrimaryPoolInLoadScript
#
#   Purpose: check for is pool primary pool in load script
#   Logic:   the assumption for this fuction is maximum no of pools will be two and
#            in case of two pools first pool will always be secondary pool
#
def NWCLSTR_IsPrimaryPoolInLoadScript(loadScriptPath, poolName):
	pools = NWCLSTR_GetPoolsFromLoadScript(loadScriptPath)
	# check if number of pools is one we will compare pooname
	if(len(pools) == 1 and pools[0] == poolName):
	    return 1;
	# check if number of pools are 2 we will comapare second pool as per logic
	elif(len(pools) == 2 and pools[1] == poolName):
	    return 1;
	else:
	    return 0;

#############################################################################
#
# Function: NWCLSTR_GetLoadScriptIfClusterEnabledPool
#
#   Purpose: will return Loadscript for only Cluster Enabled pool.
#
#   Explanation : This function will return that load script in which
#           Pool is primary pool.
#           i.e. if A1 and A2 are clustred pools.
#           while creating DST pair we have A1 as primary
#           and A2 as shadow pool
#           for A2 this function will return load script for A2
#           mostly it will be called A2_SERVER.load and not A1_SERVER.load


def NWCLSTR_GetLoadScriptIfClusterEnabledPool(poolName):
	ret= NWCLSTR_GetLoadScriptName(poolName).strip()
	scripts=ret.split('\n')

	if(ret == ""):
		return ""

	for filename in scripts:
		result= NWCLSTR_IsPrimaryPoolInLoadScript(filename,poolName)
		if (result == 1):
			return filename
	return ""

#############################################################################
#
# Function: NWCLSTR_GetNCPServerName
#
#   Purpose: getting NCP servername for pool
#

def NWCLSTR_GetNCPServerName(ld, clusterDn, args):

    name = args[1]


    loadScriptPath = NWCLSTR_GetLoadScriptIfClusterEnabledPool(name)
    if(loadScriptPath == ""):
        return ""

    try:
        awkResults = subprocess.Popen( ["/bin/sh", "-c", "awk '/ncpservername/ && !/#/' " + loadScriptPath ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
    except:
        awkResults = "";
    for awkLine in awkResults.split("\n"):
        if (awkLine != ""):
                servName = re.findall("--ncpservername=(\S+)",awkLine)
                if (len(servName)):
                    return servName[0]

    return ""

#############################################################################
#
# Function: NWCLSTR_IsClusterEnabled
#
#   Purpose:
#
def NWCLSTR_IsVolClusterEnabled(name):

    try:
        results = subprocess.Popen( ["/bin/sh","-c", "grep '\\<" + name + "\\>' /var/opt/novell/ncs/*.load |cut -d':' -f2" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0].strip()
    except:
        results = "";

    for line in results.split("\n"):
        if (line != ""):
            list = re.findall("mount\s+\w+=(\d+)",line)
            if(len(list)):
                return list[0]

    return 0


#############################################################################
#
# Function: NWCLSTR_IsClusterEnabled
#
#   Purpose:
#
def NWCLSTR_IsClusterEnabled(ld, clusterDn, args):

    name = args[1]
    objectClass = args[2]
# To find if pool is cluster enabled or not we will check NCPServerName is valid.

    if (objectClass == "nssfsPool"):
        ret = NWCLSTR_GetLoadScriptIfClusterEnabledPool(name)
	if(ret == ""):
		return 0
	else:
		return 1
# volume logic is working as before.
    try:
        results = subprocess.Popen( ["/bin/sh", "-c", "grep '\\<" + name + "\\>' /var/opt/novell/ncs/*.load" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
    except:
        results = "";

    for line in results.split("\n"):

        if (line != ""):

            fields = line.split(":")
            if (len(fields) == 2):
                scriptPath = fields[0]
                commandLine = fields[1]
                # print scriptPath + " -> " + commandLine

                commandFields = commandLine.split("=")
                if (len(commandFields) != 2):
                    continue

                if (objectClass == "volume"):
                    fields = string.split(commandFields[0])
                    if (name == fields[len(fields)-1]):
                        return 1

    return 0

#############################################################################
#
# Function: NWCLSTR_ClusterEnablePool
#
#   Purpose:
#
def NWCLSTR_ClusterEnablePool(ld, clusterDn, args):

    poolDn = eDirToLdapDn(args[1])
    ipAddress = args[2]
    serverName = string.upper(args[3])
    advertize = args[4]
    autoOnline = args[5]
    cifsServerName = args[6]

    createResourceScripts(ld, clusterDn)
    if 1 == NWCLSTR_UsedIP( ld, clusterDn, args[1:] ):
        syslog.syslog(syslog.LOG_DEBUG, "NWCLSTR_ClusterEnablePool :used ip found\n")
        return -20898

    if (syncObject(ld, poolDn) == False):
        syslog.syslog(syslog.LOG_DEBUG, "NWCLSTR_ClusterEnablePool: poolDn not found %s\n" % poolDn)
        return -1

    poolName = str(NWCLSTR_GetRawPoolName(ld, clusterDn, args))
    resourceName = poolName + "_SERVER"

    createResourcePriorityConfFile(ld, clusterDn)
    retval = clusterEnableSteps2(ld, clusterDn, poolDn, resourceName, poolName, serverName, cifsServerName, ipAddress, advertize)
    if (retval < 0):
        return (-1 + retval)

    updateResourceConfig(ld, clusterDn, resourceName)
    addResource(resourceName)

    exception = [False]
    syncResource(resourceName, "Offline", exception)
    try:
        results = subprocess.Popen( [ "cluster", "status",  resourceName ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
        if len(results) == 0:
            return (-1)
    except:
        pass

    if (autoOnline == "1"):
        try:
            pid = os.fork()
            if ( 0 == pid ):
                daemonize()
                while( not syncResource(resourceName, "Offline", exception)):
                    if(exception[0] == True):
                        syslog.syslog( syslog.LOG_ERROR, "Exception occured while getting cluster status")
                        os._exit(os.EX_SOFTWARE)
                    time.sleep( 5 )
                os.system("cluster online " + resourceName + " " + os.uname()[1])
                os._exit(0)
            else:
                os.waitpid( pid, 0 )
        except:
            pass

    return 0

#############################################################################
#
# Function: NWCLSTR_ClusterEnableVol
#
#   Purpose:
#
def NWCLSTR_ClusterEnableVol(ld, clusterDn, args):

    volumeDn = eDirToLdapDn(args[1])
    volumeName = ""
    poolDn = ""
    serverDn = ""
    resourceDn = ""
    AddScript = 0

    if (syncObject(ld, volumeDn) == False):
        return -1

    createResourceScripts(ld, clusterDn)

    try:
        res = ld.search_s(volumeDn, ldap.SCOPE_BASE)
        volumeName = res[0][1]["hostResourceName"][0]
        poolDn = res[0][1]["nssfsPool"][0]

        try:
            res = ld.search_s(poolDn, ldap.SCOPE_BASE)
            serverDn = res[0][1]["hostServer"][0]

            try:
                res = ld.search_s(serverDn, ldap.SCOPE_BASE)
                resourceDn = res[0][1]["Resource"][0]

            except:
                return -5

        except:
            return -4

    except:
        return -3

    if 2 < len( args ):
        if (args[2] == 0):
            volumeId = NWCLSTR_IsVolClusterEnabled( volumeName)
            if (0 != volumeId):
                AddScript = 1
        else:
            volumeId = int( args[2] )
    else:
        volumeId = 0

    if 0 == volumeId:
        volumeId = generateVolumeId()

    if (volumeId == -1):
        return -2

    clusterVolumeDn = "cn=" + objectNameFromLdapDn(clusterDn) + "_" + volumeName
    for component in string.split(volumeDn, ",")[1:]:
        clusterVolumeDn += "," + component

    try:
        ld.rename_s(volumeDn, clusterVolumeDn)

    except:
        return -6

    attrs = []
    attrs.append( (ldap.MOD_ADD, "nCSVolumes", clusterVolumeDn) )

    try:
        ld.modify_s(serverDn, attrs)

    except:
        return -7

    try:
        ld.modify_s(resourceDn, attrs)

    except:
        return -8

    attrs = []
    attrs.append( (ldap.MOD_REPLACE, "hostServer", serverDn) )

    try:
        ld.modify_s(clusterVolumeDn, attrs)

    except:
        return -9
    if (0 == AddScript):
        retval, revision = addVolumeToResourceScripts(ld, clusterDn, resourceDn, volumeName, volumeId)
        if (retval < 0):
            return (-9 + retval)

    try:
        pid = os.fork()
        if ( 0 == pid ):
            daemonize()
            updateResource(objectNameFromLdapDn(resourceDn), revision)
            os.system("/sbin/ncpcon mount " + volumeName + "=" + str(volumeId))
            os._exit(0)
        else:
            os.waitpid( pid, 0 )

    except:
        pass

    return 0

#############################################################################
#
# Function: NWCLSTR_EnableAllVolumesOnPool
#
#   Purpose:
#
def NWCLSTR_EnableAllVolumesOnPool(ld, clusterDn, args):

    poolDn = eDirToLdapDn(args[1])
    ipAddress = args[2]
    serverName = string.upper(args[3])
    cifsServerName = args[4]
    advertize = args[5]

    if  1 == NWCLSTR_UsedIP( ld, clusterDn, args[1:] ):
        return -20898

    if (syncObject(ld, poolDn) == False):
        return -1

    poolName = NWCLSTR_GetRawPoolName(ld, clusterDn, args)
    resourceName = poolName + "_SERVER"

    clusterPoolDn = "cn=" + objectNameFromLdapDn(clusterDn) + "_" + poolName + "_POOL"
    for component in string.split(poolDn, ",")[1:]:
        clusterPoolDn += "," + component

    retval = clusterEnableSteps2(ld, clusterDn, poolDn, resourceName, poolName, serverName, cifsServerName, ipAddress, advertize)
    if (retval < 0):
        return (-1 + retval)

    updateResourceConfig(ld, clusterDn, resourceName)

    try:
        res = ld.search_s(containerFromLdapDn(poolDn), ldap.SCOPE_ONELEVEL, "(&(objectClass=volume)(nssfsPool=" + poolDn + "))", [ "nssfsPool" ] )

        for volume in res:
            if volume[1].has_key("nssfsPool"):
                if (poolDn == volume[1]["nssfsPool"][0]):
                    NWCLSTR_ClusterEnableVol(ld, clusterDn, [ None, volume[0] ])

    except:
        pass

    try:
        ld.rename_s(poolDn, clusterPoolDn)

    except:
        pass

    return resourceName

#############################################################################
#
# Function: getResourceNameFromPoolName
#
#   Purpose:
#
def getResourceNameFromPoolName(poolName):

    try:
        return subprocess.Popen( [ "/bin/sh", "-c", "grep -e \"nss[[:space:]]\\+\\/poolact[^\>]*\\>=\\<" + poolName + "\\>\" /var/opt/novell/ncs/*.load | grep -o -e \"[^\\/]*\\.load\" | sed -e \"s/\\.load$//\" | tr -d \"\\n\"" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
    except:
        return ""

#############################################################################
#
# Function: NWCLSTR_CIFSVolServ
#
#   Purpose:
#
def NWCLSTR_CIFSVolServ(ld, clusterDn, args):

    try:
        res = ld.search_s(args[1], ldap.SCOPE_BASE)
    except:
        return "none"

    if res[0][1].has_key( "nfapCIFSServerName" ):
        return res[0][1]["nfapCIFSServerName"][0]

    return "none"

#############################################################################
#
# Function: NWCLSTR_PoolRename
#
#   Purpose:
#
def NWCLSTR_PoolRename(ld, clusterDn, args):

    if(NWCLSTR_SafeToUpgradeFS("OES2018SP3") == 0):
        EDIR_ATTR_LIMIT = CLSTRCONF_SCRIPTS_BUFFER_LEN
    else:
       EDIR_ATTR_LIMIT = EDIR_ATTR_LIMIT_OLD

    poolName = args[1]
    newName = args[2]

    resourceName = getResourceNameFromPoolName(poolName)

    if ("" != resourceName):
#        os.system("cluster offline " + resourceName)

        resourceDn = "cn=" + resourceName + "," + clusterDn;

        try:
            res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        except:
            return -2

        loadScript = res[0][1]["nCSCRMLoadScript"][0]
        unloadScript = res[0][1]["nCSCRMUnloadScript"][0]
        monitorScript = res[0][1]["nCSCRMMonitorScript"][0]
        revision = int(res[0][1]["nCSRevision"][0])

        newRevision = revision + 1
        newLoadScript = re.sub( r"(nss\s+/poolact[^=]*=)" + poolName + r"(\s)", r"\g<1>" + newName + r"\2", loadScript)
        newUnloadScript = re.sub( r"(nss\s+/pooldeact[^=]*=)" + poolName + r"(\s)", r"\g<1>" + newName + r"\2", unloadScript)
        newMonitorScript = re.sub( r"/" + poolName + r"\b", r"/" + newName, monitorScript)

        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSRevision", str(newRevision)) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadScript", newLoadScript[:EDIR_ATTR_LIMIT]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMUnloadScript", newUnloadScript[:EDIR_ATTR_LIMIT]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorScript", newMonitorScript[:EDIR_ATTR_LIMIT]) )

        try:
            ld.modify_s(resourceDn, attrs)
        except:
            return -3

        try:
            updateResourceConfig(ld, clusterDn, resourceName)
            pid = os.fork()
            if ( 0 == pid ):
                daemonize()
                os._exit( os.system( "sleep 12 && echo -n 'exec /opt/novell/ncs/bin/ncs-configd.py -init &' > /proc/ncs/cluster" ))
            else:
                os.waitpid( pid, 0 )
        except:
            return -4

    return 0

#############################################################################
#
# Function: NWCLSTR_PoolDelete
#
#   Purpose: Delete a resource that is associated with a deleted pool
#
#   Note: The resource must be offline otherwise this function will fail
def NWCLSTR_PoolDelete(ld, clusterDn, args):

    poolName = args[1]
    resourceName = getResourceNameFromPoolName(poolName)
    if resourceName == None:
        return 0

    try:
        pid = os.fork()
        if ( 0 == pid ):
            daemonize()
            xml = '<virtualIO><datastream name=\"command\"/></virtualIO>'
            xml += '<ncsRequest><clusterConfig><deleteClusterResource>'
            xml += '<name>' + resourceName + '</name>'
            xml += '</deleteClusterResource></clusterConfig></ncsRequest>'
            syslog.syslog(syslog.LOG_INFO, xml );

            try:
                fd = os.open("/admin/Novell/Cluster/ClusterConfig.cmd", os.O_RDWR)
                os.write(fd, xml)
                os.lseek(fd, 0, 0)
                results = os.read(fd, 512)
                os.close(fd)
                syslog.syslog(syslog.LOG_INFO, results );
            except:
                syslog.syslog(syslog.LOG_ERR, "NWCLSTR_PoolDelete: adminfs IO failed\n")

            os._exit(0)
        else:
            os.waitpid( pid, 0 )

    except:
         syslog.syslog(syslog.LOG_ERR, "NWCLSTR_PoolDelete: fork error\n")

    #Todo: return result value parsed from the results xml
    return 0

#############################################################################
#
# Function: getResourceNameFromVolumeName
#
#   Purpose:
#
def getResourceNameFromVolumeName(volumeName):

    try:
        return subprocess.Popen( ["/bin/sh", "-c", "grep -e \"ncpcon[[:space:]]\\+mount[[:space:]]\\+\\<" + volumeName + "\\>\" /var/opt/novell/ncs/*.load | grep -o -e \"[^\\/]*\\.load\" | sed -e \"s/\\.load$//\" | tr -d \"\\n\"" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
    except:
        return ""

#############################################################################
#
# Function: NWCLSTR_RenameResource
#
#   Purpose:
#
def NWCLSTR_RenameResource(ld, clusterDn, args):

    oldResourceName = "cn=" + args[1] + "," + clusterDn
    newResourceName = "cn=" + args[2] + "," + clusterDn

    if not syncObject( ld, oldResourceName ):
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: resource " + args[1] + " is not found.\n" )
        return -1

    if os.system( "cluster status \"" + args[1] + "\" | grep -e '^Offline[[:space:]]' &> /dev/null" ):
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: resource " + oldResourceName + " is not in offline state.\n" )
        return -2

    try:
        result = ld.search_s( oldResourceName, ldap.SCOPE_BASE, "cn=" + args[1], [ "nCSRevision" ] )
        resource_rev = result[0][1]["nCSRevision"][0]
    except Exception, e:
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: resource " + oldResourceName + " is not found in eDir.\n" )
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: %s" % (str(e)))
        return -3

    try:
        ld.rename_s( oldResourceName, newResourceName )
    except Exception, e:
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: Failed to rename " + oldResourceName + " to " + newResourceName + ".\n" )
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: %s" % (str(e)))
        return -4

    try:
        result = ld.search_s( clusterDn, ldap.SCOPE_BASE )
        stream = result[0][1]["nCSCRMResourcePriority"][0]
        priority_rev = int(  re.match( "\d+", stream[9:]).group(0))
        attrs = []
        attrs.append(( ldap.MOD_REPLACE, "nCSCRMResourcePriority", stream.replace( "=" + args[1] + "=",  "=" + args[2] + "=" ).replace(  "REVISION=" + str( priority_rev ), "REVISION=" + str( priority_rev + 1 ))))
        ld.modify_s( clusterDn, attrs )
    except Exception, e:
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: Failed to modify resource priority list.\n" )
        syslog.syslog(syslog.LOG_ERR, "NWCLSTR_RenameResource: %s" % (str(e)))
        return -5

    updateClusterConfig( ld, clusterDn )

    os.system( "echo -n 'RESOURCES " + args[2] + " =" + resource_rev + "' > /proc/ncs/cluster" )
    os.system( "echo -n 'REVISION " + str( priority_rev + 1 ) + "' > /proc/ncs/cluster" )

    return 0

#############################################################################
#
# Function: NWCLSTR_VolRename
#
#   Purpose:
#
def NWCLSTR_VolRename(ld, clusterDn, args):

    if(NWCLSTR_SafeToUpgradeFS("OES2018SP3") == 0):
        EDIR_ATTR_LIMIT = CLSTRCONF_SCRIPTS_BUFFER_LEN
    else:
        EDIR_ATTR_LIMIT = EDIR_ATTR_LIMIT_OLD

    volumeName = args[1]
    newName = args[2]

    resourceName = getResourceNameFromVolumeName(volumeName)

    if ("" != resourceName):
        resourceDn = "cn=" + resourceName + "," + clusterDn;

        try:
            res = ld.search_s(resourceDn, ldap.SCOPE_BASE)
        except:
            return -2

        loadScript = res[0][1]["nCSCRMLoadScript"][0]
        monitorScript = res[0][1]["nCSCRMMonitorScript"][0]
        revision = int(res[0][1]["nCSRevision"][0])

        newRevision = revision + 1
        newLoadScript = re.sub( r"(\s)" + volumeName + r"=", r"\g<1>" + newName + r"=", loadScript)
        newMonitorScript = re.sub( r"\b" + volumeName + r"\b", newName, monitorScript)

        attrs = []
        attrs.append( (ldap.MOD_REPLACE, "nCSRevision", str(newRevision)) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadScript", newLoadScript[:EDIR_ATTR_LIMIT]) )
        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorScript", newMonitorScript[:EDIR_ATTR_LIMIT]) )

        try:
            ld.modify_s(resourceDn, attrs)
        except:
            return -3

        try:
            updateResourceConfig(ld, clusterDn, resourceName)
        except:
            return -4

        try:
            pid = os.fork()
            if ( 0 == pid ):
                daemonize()
                updateResource( resourceName, newRevision )
                matchCMD = re.search( "^\s*exit_on_error\s+(ncpcon\s+mount\s+" + newName + "=.*)$", newLoadScript, re.M )
                if matchCMD:
                    os.system( matchCMD.group( 1 ))
                os._exit(0)
            else:
                os.waitpid( pid, 0 )

        except:
            pass

    return 0

#############################################################################
#
# Function: NWCLSTR_VolDelete
#
#   Purpose:
#
def NWCLSTR_VolDelete(ld, clusterDn, args):

    retval = 0;
    volumeName = args[1]
    resourceName = getResourceNameFromVolumeName(volumeName)

    if ("" != resourceName):
        retval, revision = removeVolumeFromResourceScripts(ld, clusterDn, "cn=" + resourceName + "," + clusterDn, volumeName)

    return retval

#############################################################################
#
# Function: NWCLSTR_PreviewScriptTranslation
#   Purpose: Previews the load/unload script translation that is done via
#               the function translateNetWareScript().
#
def NWCLSTR_PreviewScriptTranslation(ld, clusterDn, args):

    results = ""
    resourceName = args[1];

    try:
        resourceDn = "cn=" + resourceName + "," + clusterDn
        res = ld.search_s(resourceDn, ldap.SCOPE_BASE)

        resource = res[0]

        if resource[1].has_key("nCSCRMLoadScript"):
            changed, script = translateNetWareScript(resource[1]["nCSCRMLoadScript"][0])

            if (changed):
                results += "## Load script (with modifications) ##\n"
            else:
                results += "## Load script (no modifications necessary) ##\n"

            results += script + "\n"
            results += "## Monitor script ##\n"
            results += generateMonitorScript( script ) + "\n"

        if resource[1].has_key("nCSCRMUnloadScript"):
            changed, script = translateNetWareScript(resource[1]["nCSCRMUnloadScript"][0])

            if (changed):
                results += "## Unload script (with modifications) ##\n"
            else:
                results += "## Unload script (no modifications necessary) ##\n"

            results += script

    except ldap.NO_SUCH_OBJECT:
        results = -1

    except ldap.LDAPError:
        results = -2

    except:
        results = -1000

    return(results)

#############################################################################
#
# Function: NWCLSTR_PreviewAllScripts
#   Purpose: Preview translations for all the resources
#
def NWCLSTR_PreviewAllScripts(ld, clusterDn, args):

    results = ""

    res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "(|(objectClass=nCSClusterResource)(objectClass=nCSVolumeResource))", [ "cn" ] )
    for resource in res:
        results += "\nFor resoruce " +  resource[1]['cn'][0] + ":\n\n"
        results += str( NWCLSTR_PreviewScriptTranslation( ld, clusterDn,  [ "", resource[1]['cn'][0] ]))
        results += "\n"

    return(results)

#############################################################################
#
# Function: NWCLSTR_ConvertCluster
#   Purpose: Converts a mixed cluster to a pure Linux cluster
#
def NWCLSTR_ConvertCluster(ld, clusterDn, args):

    retVal = 0

    try:
        # Check cluster revision.  This will only convert a cluster that is at
        # revision 280 or 281.
        res = ld.search_s(clusterDn, ldap.SCOPE_BASE)
        revision = int(res[0][1]["nCSRevision"][0])
        if (revision < 280):
            retVal = -1         # Cannot convert a pure NetWare cluster
        elif (revision > 281):
            retVal = -255       # Cluster is already converted
            res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "(|(objectClass=nCSClusterResource)(objectClass=nCSVolumeResource))" )
            for resource in res:
                attrs = []
                syslog.syslog(syslog.LOG_INFO, "Converting %d res : processing : %s " %(revision, str(resource[1]["cn"][0])))

                if resource[1].has_key("nCSCRMLoadScript"):
                    if not resource[1].has_key("nCSCRMMonitorScript"):
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorScript", generateMonitorScript( resource[1]["nCSCRMLoadScript"][0] )) )
                    if not resource[1].has_key("nCSCRMMonitorTimeout"):
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorTimeout", "360") )
                    if not resource[1].has_key("nCSCRMMonitorFailurePeriod"):
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorFailurePeriod", "600") )
                    if not resource[1].has_key("nCSCRMMonitorFrequency"):
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorFrequency", "60") )
                    if not resource[1].has_key("nCSCRMMonitorMaxFailures"):
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorMaxFailures", "3") )
                if (len(attrs) > 0):
                    attrs.append( (ldap.MOD_REPLACE, "nCSRevision", str(int(resource[1]["nCSRevision"][0]) + 1)) )
                    ld.modify_s(resource[0], attrs)
                    retVal = 0
        else:
            res = ld.search_s(clusterDn, ldap.SCOPE_ONELEVEL, "(|(objectClass=nCSClusterResource)(objectClass=nCSVolumeResource))" )

            for resource in res:

                attrs = []
                syslog.syslog(syslog.LOG_INFO, "Converting %d -> 282 : processing : %s " %(revision, str(resource[1]["cn"][0])))

                if resource[1].has_key("nCSCRMLoadScript"):
                    changed, script = translateNetWareScript(resource[1]["nCSCRMLoadScript"][0])

                    if (changed):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMLoadScript", script) )

                    if resource[1].has_key("nCSCRMMonitorScript"):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorScript", generateMonitorScript( script )) )
                    else:
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorScript", generateMonitorScript( script )) )

                    if resource[1].has_key("nCSCRMMonitorTimeout"):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorTimeout", "360") )
                    else:
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorTimeout", "360") )

                    if resource[1].has_key("nCSCRMMonitorFailurePeriod"):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorFailurePeriod", "600"))
                    else:
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorFailurePeriod", "600"))

                    if resource[1].has_key("nCSCRMMonitorFrequency"):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorFrequency", "60") )
                    else:
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorFrequency", "60") )

                    if resource[1].has_key("nCSCRMMonitorMaxFailures"):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMMonitorMaxFailures", "3") )
                    else:
                        attrs.append( (ldap.MOD_ADD, "nCSCRMMonitorMaxFailures", "3") )

                if resource[1].has_key("nCSCRMUnloadScript"):
                    changed, script = translateNetWareScript(resource[1]["nCSCRMUnloadScript"][0])

                    if (changed):
                        attrs.append( (ldap.MOD_REPLACE, "nCSCRMUnloadScript", script) )

                if (len(attrs) > 0):
                    if resource[1].has_key("nCSRevision"):
                        attrs.append( (ldap.MOD_REPLACE, "nCSRevision", str(int(resource[1]["nCSRevision"][0]) + 1)) )
                    else:
                        attrs.append( (ldap.MOD_ADD, "nCSRevision", "1") )
                    ld.modify_s(resource[0], attrs)

            attrs = []
            attrs.append( (ldap.MOD_REPLACE, "nCSRevision", "282") )
            ld.modify_s(clusterDn, attrs)

            syslog.syslog(syslog.LOG_INFO, "The cluster has been converted to version 282")

            import ncstempl
            ncstempl.updateResourceTemplates(ld, clusterDn)
            createResourcesXmlFile(ld, clusterDn)
            createResourceScripts(ld, clusterDn)

            syslog.syslog(syslog.LOG_INFO, "The cluster resource templates have been updated")

    except ldap.LDAPError:
        retVal = -2

    except:
        retVal = -1000

    return(retVal)

#############################################################################
#
# Function: NWCLSTR_UpdateClusterConfig
#   Purpose:
#
def NWCLSTR_UpdateClusterConfig(ld, clusterDn, args):
    updateClusterConfig(ld, clusterDn)
    return 0

#############################################################################
#
# Function: NWCLSTR_ReadConfFile
#   Purpose:
#
def NWCLSTR_ReadConfFile(ld, clusterDn, args):
    content = ""
    for i in range( 1, 5 ):
        try:
            fd = open( args[1], 'r' )
            content = fd.read()
            fd.close()
            break
        except:
            time.sleep( 1 )
            content = ""

    return content

#############################################################################
#
# Function: NWCLSTR_GetPoolVersion
#   Purpose:
#
def NWCLSTR_GetPoolVersion(ld, clusterDn, args):

    try:
        load_file_name = clstrlibPath + "/" + args[1] + ".load"
        if not os.path.isfile( load_file_name ):
            os.system( "/opt/novell/ncs/bin/ncs-configd.py -init &>/dev/null" )

        rc = subprocess.Popen( [ "/bin/sh", "-c", "grep -io '^[^#]*nss /poolact=[^[:space:]]\+' " + load_file_name + " 2>/dev/null | grep -o '[^=]\+$'" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]
        while '$' == rc[0]:
            matchVar=re.search( '^\${?([^}\s]+)}?', rc )
            if matchVar and matchVar.group(1):
                rc = subprocess.Popen( [ "/bin/sh", "-c", "grep -o '^[^#$]*" + matchVar.group(1) + "=[^[:space:]]\+' " + load_file_name + " 2>/dev/null | cut -d'=' -f2 | sed 's/^[\\\"'\\\'']//' | sed 's/[\\\"'\\\'']$//'" ], close_fds=True, stdout=subprocess.PIPE ).communicate()[0]

        return rc.upper()
    except:
        return ""

#############################################################################
#
# Function: NWCLSTR_Repair
#   Purpose:
#
def NWCLSTR_Repair(ld, clusterDn, args):
    rc = 0

    try:
        pid = os.fork()
        if( 0 == pid ):
            daemonize()
            os._exit( os.system( "(date; /opt/novell/ncs/bin/ncs_repair.py; date; echo) 2>&1 >>/var/opt/novell/log/ncs/repair.log </dev/null" ))
        else:
            os.waitpid( pid, 0 )
    except:
        rc = 1

    return rc

def NWCLSTR_SafeToUpgradeFS(version):
    rc = 0
    if version:
        if(version == "OES2015"):
            tag = "0x29"
        elif version == "OES-SP1":
            tag = "0x2a"
        elif version == "OES2015SP1":
            tag = "0x28"
        elif version == "OES2018":
            tag = "0x27"
        elif version == "OES2018SP2":
            tag = "0x26"
        elif version == "OES2018SP3":
            tag = "0x25"
    child = subprocess.Popen("grep '^nodeid' /var/opt/novell/ncs/gipc.conf | wc -l", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    no_of_nodes = child.communicate()[0]
    child = subprocess.Popen(["/opt/novell/ncs/bin/sbdutil", "-q", no_of_nodes, "-t", tag], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    child.wait()
    rc = child.returncode

    return rc

